// Main private header file for VX environment library
#ifndef ENV_H
#define ENV_H

#include <sys/types.h>
#include <inttypes.h>

#include "vx/vx.h"	// Public definitions


// 8-bit registers, numbered according to x86 encodings
#define AL	0
#define CL	1
#define DL	2
#define BL	3
#define AH	4
#define CH	5
#define DH	6
#define BH	7

// 16-bit registers
#define AX	0
#define CX	1
#define DX	2
#define BX	3
#define SP	4
#define BP	5
#define SI	6
#define DI	7

// 32-bit registers
#define EAX	0
#define ECX	1
#define EDX	2
#define EBX	3
#define ESP	4
#define EBP	5
#define ESI	6
#define EDI	7


// EFLAGS bits implemented in vx32
#define EFL_CF_BIT	0
#define EFL_PF_BIT	2
#define EFL_ZF_BIT	6
#define EFL_SF_BIT	7
#define EFL_DF_BIT	10
#define EFL_OF_BIT	11
#define EFL_CF		(1 << EFL_CF_BIT)
#define EFL_PF		(1 << EFL_PF_BIT)
#define EFL_ZF		(1 << EFL_ZF_BIT)
#define EFL_SF		(1 << EFL_SF_BIT)
#define EFL_DF		(1 << EFL_DF_BIT)
#define EFL_OF		(1 << EFL_OF_BIT)


// 16-bit condition code values used in Jcc, SETcc, CMOVcc instructions
#define CC_O	0x0	// Overflow
#define CC_NO	0x1	// Not Overflow
#define CC_B	0x2	// Below
#define CC_NAE	0x2	// Not Above or Equal
#define CC_C	0x2	// Carry
#define CC_AE	0x3	// Above or Equal
#define CC_NB	0x3	// Not Below
#define CC_NC	0x3	// Not Carry
#define CC_E	0x4	// Equal
#define CC_Z	0x4	// Zero
#define CC_NE	0x5	// Not Equal
#define CC_NZ	0x5	// Not Zero
#define CC_BE	0x6	// Below or Equal
#define CC_NA	0x6	// Not Above
#define CC_A	0x7	// Above
#define CC_NBE	0x7	// Not Below or Equal
#define CC_S	0x8	// Sign
#define CC_NS	0x9	// Not Sign
#define CC_P	0xa	// Parity
#define CC_PE	0xa	// Parity Even
#define CC_NP	0xb	// Not Parity
#define CC_PO	0xb	// Parity Odd
#define CC_L	0xc	// Less
#define CC_NGE	0xc	// Not Greater or Equal
#define CC_GE	0xd	// Greater or Equal
#define CC_NL	0xd	// Not Less
#define CC_LE	0xe	// Less or Equal
#define CC_NG	0xe	// Not Greater
#define CC_G	0xf	// Greater
#define CC_NLE	0xf	// Not Less or Equal


struct vxproc {

	// Index of this vxproc in vxprocs array (see below).
	// Also used to compute our code and data segment selectors.
	uint32_t	vxpno;

	// File descriptor for the vx32 process's address space.
	int		fd;

	// Current size of the process's address space.
	// Keep this a multiple of the page size (4KB)!
	uint32_t	size;

	// Data segment for the vx32 process, NULL if not currently mapped.
	// Mapping size is always the same as address space size ('size').
	void		*data;

	// Permission table for the address space.
	// Table size is one byte per page of address size (size/4096).
	uint8_t		*perm;

	// Host architecture-specific emulation state for this process
	struct vxemu	*emu;
};


// Parameters tweakable for performance
#define VXPROCSMAX	256		// Max # of vxprocs per host proc
#define VXCODEALIGN	16		// Alignment for translated code
#define VXCODEBUFSIZE	(1024*1024)	// Code buffer size (XX make variabale)
#define VXENTRYHASHLEN	32768		// # entrypoints in entry hash table


// vx32 Environment management functions private to libvx.

int vxproc_setupseg(struct vxproc *vxp);
int vxproc_map(struct vxproc *proc);
int vxproc_resize(struct vxproc *proc, size_t newsize);
void vxproc_unmap(struct vxproc *proc);
void vxproc_clear(struct vxproc *proc);


// Architecture- and platform-specific emulation code
// must implement the following instructions.

int vxemu_init(struct vxproc *proc);
void vxemu_free(struct vxemu *e);
void vxemu_flush(struct vxemu *e);
int vxproc_run(struct vxproc *proc);
int vxemu_map(struct vxemu *e);


#endif // ENV_H
