// VX library functions for Unix hosts

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <ucontext.h>

#include "vx/vx.h"
#include "vx/vx32.h"
#include "vx/proc.h"

#include "env.h"
#include "emu.h"


// Host-specific modules define this if they need any specific flags
// to be set when we mmap a vx process's memory.
#ifndef VX_MMAP_FLAGS
#define VX_MMAP_FLAGS	0
#endif


static vxproc *procs[VXPROCSMAX];


vxproc *vxproc_new(void)
{
	// Find an available process number
	int pno;
	for (pno = 0; ; pno++) {
		if (pno == VXPROCSMAX) {
			errno = EAGAIN;
			return NULL;
		}
		if (procs[pno] == 0)
			break;
	}

	// Allocate the vxproc structure
	vxproc *pr = calloc(1, sizeof(vxproc));
	if (pr == NULL) {
		errno = ENOMEM;
		return NULL;
	}
	pr->vxpno = pno;
	pr->fd = -1;		// Address space starts out empty
	pr->size = 0;
	pr->data = NULL;
	pr->perm = NULL;

	// Create the process's emulation state
	if (vxemu_init(pr) < 0)
		return NULL;

	procs[pno] = pr;
	return pr;
}

// Unmap a vxproc's data segment from our address space.
// This doesn't cause us to lose its actual contents,
// since we still have the open file handle it maps.
void vxproc_unmap(vxproc *proc)
{
	if (proc->data != NULL) {
		munmap(proc->data, proc->size);
		proc->data = NULL;
	}
}

// Clear a vxproc's entire address space to zero.
// We do this simply by closing the file representing the address space;
// a new file will be created next time it is needed.
void vxproc_clear(vxproc *proc)
{
	vxproc_unmap(proc);

	if (proc->fd >= 0) {
		close(proc->fd);
		proc->fd = -1;
	}
	proc->size = 0;
}

// Convert a VX permission value into a PROT_* value for POSIX mprotect().
// Also check the VX permission for legality,
// returning -1 and errno=EINVAL on illegal values.
static int getprot(int perm)
{
	switch (perm) {
	case 0:
		return 0;

	case VXPERM_READ:
		return PROT_READ;

	case VXPERM_READ | VXPERM_WRITE:
		return PROT_READ | PROT_WRITE;

	case VXPERM_READ | VXPERM_EXEC:
		// Never include exec permission in our host mapping
		// since we should never be executing the env's code directly.
		return PROT_READ;

	default:
		errno = EINVAL;
		return -1;
	}
}

// Create the host file representing a process's address space.
int vxproc_initspace(vxproc *proc)
{
	if (proc->fd >= 0)
		return 0;

	// Create a temporary file to hold the process's address space
	char tmpfn[] = "/tmp/vxXXXXXX";
	int fd = mkstemp(tmpfn);
	if (fd < 0)
		return -1;

	// Immediately remove it now that we have it open.
	// That way it will disappear as soon as we close the handle.
	(void)remove(tmpfn);

	proc->fd = fd;
	return 0;
}

// Map a process's data segment into directly-accessible memory.
int vxproc_map(vxproc *proc)
{
	if (proc->data != NULL)
		return 0;

	// Make sure the address space is initialized.
	if (vxproc_initspace(proc) < 0)
		return -1;

	// Map the process's address space file into our memory,
	// initially with no access permissions anywhere in the region.
	void *p = mmap(NULL, proc->size, PROT_NONE,
			MAP_SHARED | VX_MMAP_FLAGS, proc->fd, 0);
	if (p == MAP_FAILED)
		return -1;

	// Now set the permissions on all pages that should be accessible.
	unsigned npages = proc->size / VXPAGESIZE;
	for (unsigned i = 0; i < npages; ) {

		uint8_t pageperm = proc->perm[i];

		// For efficiency, merge mprotect calls on page ranges.
		unsigned j;
		for (j = i+1; j < npages; j++)
			if (proc->perm[j] != pageperm)
				break;

		if (pageperm == 0) {
			i = j;		// nothing to do for this range
			continue;
		}

		// Calculate the effective permission to map with in the host.
		int prot = getprot(pageperm);
		assert(prot >= 0);

		// Set the permissions on this range.
		if (mprotect(p + i*VXPAGESIZE, (j-i)*VXPAGESIZE, prot) < 0) {
			munmap(p, proc->size);
			return -1;
		}

		i = j;
	}

	// Record the mapped data segment
	proc->data = p;

	// Perform any emulator-specific mapping setup
	if (vxemu_map(proc->emu) < 0) {
		vxproc_unmap(proc);
		return -1;
	}

	return 0;
}

int vxproc_resize(vxproc *proc, size_t newsize)
{
	// Make sure the address space is initialized.
	if (vxproc_initspace(proc) < 0)
		return -1;

	assert((newsize & (VXPAGESIZE-1)) == 0);
	if (proc->size == newsize)
		return 0;

	// Any mapping we may have will not be valid for the new size.
	vxproc_unmap(proc);

	// Allocate a new permissions array if the process is growing.
	uint8_t *newperm = NULL;
	if (newsize > proc->size) {
		size_t oldpages = proc->size / VXPAGESIZE;
		size_t newpages = newsize / VXPAGESIZE;
		newperm = malloc(newpages);
		if (newperm == NULL) {
			errno = ENOMEM;
			return -1;
		}
		if (oldpages > 0)
			memcpy(newperm, proc->perm, oldpages);
		memset(newperm+oldpages, 0, newpages-oldpages);
	}

	// Grow or truncate the file to the requested size
	if (ftruncate(proc->fd, newsize) < 0) {
		if (newperm != NULL)
			free(newperm);
		return -1;
	}

	proc->size = newsize;
	if (newperm != NULL) {
		if (proc->perm != NULL)
			free(proc->perm);
		proc->perm = newperm;
	}

	return 0;
}

int vxproc_setperm(vxproc *proc, intptr_t addr, size_t size, int perm)
{
	// Validate the region specified
	size_t lo = addr;
	size_t hi = addr + size;
	if (hi < lo || hi > VXMEMSIZE) {
		errno = EINVAL;
		return -1;
	}
	lo = VXPAGETRUNC(lo);
	hi = VXPAGEROUND(hi);

	// Validate the permissions specified
	int prot = getprot(perm);
	if (prot < 0)
		return -1;

	// Make sure the process is big enough
	if (proc->size < hi) {
		// Give the process 1MB at a time
		// so that we don't have to re-mmap()
		// for each new page it allocates.
		size_t newsize = (hi + 1024*1024-1) & ~(1024*1024-1);
		vxproc_resize(proc, newsize);
	}

	// Set the permissions in the perm array.
	// If we're removing execute permission from any pages,
	// then we'll have to flush the emulation state.
	if (!(perm & VXPERM_EXEC)) {
		int flushemu = 0;
		size_t lopage = lo >> VXPAGESHIFT;
		size_t hipage = hi >> VXPAGESHIFT;
		for (size_t i = lopage; i < hipage; i++) {
			if (proc->perm[i] & VXPERM_EXEC)
				flushemu = 1;
			proc->perm[i] = perm;
		}
		if (flushemu)
			vxemu_flush(proc->emu);
	} else
		memset(&proc->perm[lo/VXPAGESIZE], perm, (hi-lo)/VXPAGESIZE);

	// Update the permissions in our host mapping of the address space
	if (proc->data != NULL) {
		if (mprotect(proc->data + lo, hi - lo, prot) < 0) {
			munmap(proc->data, proc->size);
			return -1;
		}
	}

	return 0;
}

int vxproc_checkperm(vxproc *p, size_t va, size_t len, int perm,
		size_t *out_faultva)
{
	if (len == 0)
		return 1;

	// Check the page containing the first byte of the range
	if (va >= p->size || (p->perm[va / VXPAGESIZE] & perm) != perm) {
		if (out_faultva != NULL)
			*out_faultva = va;	// First byte of range
		return 0;
	}

	// Check any remaining pages
	size_t pva = (va + VXPAGESIZE) & ~(VXPAGESIZE-1);
	while ((pva - va) < len) {
		if (pva >= p->size ||
				(p->perm[pva / VXPAGESIZE] & perm) != perm) {
			if (out_faultva != NULL)
				*out_faultva = pva;	// First byte of page
			return 0;
		}
		pva += VXPAGESIZE;
	}

	// Permission granted
	return 1;
}

void *vxproc_maprange(vxproc *p, size_t va, size_t len, int perm)
{
	if (vxproc_map(p) < 0)
		return NULL;

	if (!vxproc_checkperm(p, va, len, perm, NULL)) {
		errno = EACCES;
		return NULL;
	}

	return p->data + va;
}

void vxproc_unmaprange(vxproc *p)
{
	// no-op for now
}

int vxproc_write(vxproc *proc, const void *data, intptr_t addr, size_t size)
{
	if (addr > VXMEMSIZE || size > VXMEMSIZE || addr + size > VXMEMSIZE) {
		errno = EINVAL;
		return -1;
	}

	if (addr + size > proc->size) {
		if (vxproc_resize(proc, addr + size) < 0)
			return -1;
	}

	if (lseek(proc->fd, addr, SEEK_SET) < 0)
		return -1;
	if (write(proc->fd, data, size) < 0)
		return -1;

	return 0;
}

int vxproc_read(vxproc *proc, intptr_t addr, void *buf, size_t size)
{
	if (addr > VXMEMSIZE || size > VXMEMSIZE || addr + size > VXMEMSIZE) {
		errno = EINVAL;
		return -1;
	}

	if (addr + size > proc->size) {
		if (vxproc_resize(proc, addr + size) < 0)
			return -1;
	}

	if (lseek(proc->fd, addr, SEEK_SET) < 0)
		return -1;
	if (read(proc->fd, buf, size) < 0)
		return -1;

	return 0;
}


int vxproc_load(vxproc *proc,
		ssize_t (*readcb)(void *data, off_t ofs,
				void *buf, size_t size),
		void *cbdata,
		const char *const argv [], const char *const envp [])
{
	return vx32_load(proc, readcb, cbdata, argv, envp);
}


// Loading a VX executable from host memory

struct loadmemdesc {
	const void *buf;
	size_t size;
};

static ssize_t loadmemcb(void *cbdata, off_t ofs, void *buf, size_t size)
{
	struct loadmemdesc *d = cbdata;
	if (ofs >= d->size)
		return 0;
	if (ofs + size > d->size)
		return 0;	// no partial reads
	memcpy(buf, d->buf + ofs, size);
	return size;
}

int vxproc_loadmem(vxproc *proc, const void *exe, size_t exesize,
		const char *const argv [], const char *const envp [])
{
	struct loadmemdesc d = {exe, exesize};
	return vxproc_load(proc, loadmemcb, &d, argv, envp);
}


// Loading a VX executable from a host file

static ssize_t loadfilecb(void *cbdata, off_t ofs, void *buf, size_t size)
{
	int fd = (int)(intptr_t)cbdata;
	if (lseek(fd, ofs, SEEK_SET) < 0)
		return -1;
	return read(fd, buf, size);
}

int vxproc_loadfile(vxproc *proc, const char *filename,
		const char *const argv [], const char *const envp [])
{
	int fd = open(filename, O_RDONLY);
	if (fd < 0)
		return -1;

	if (vxproc_load(proc, loadfilecb, (void*)(intptr_t)fd, argv, envp) < 0)
		return -1;

	close(fd);
	return 0;
}


static void sighandler(int signo, siginfo_t *si, void *ucontext)
{
	ucontext_t *uctx = ucontext;
	if (!vxproc_sighandler(signo, &uctx->uc_mcontext)) {

		// Not handled - just switch back to the default handler
		// and let the signal happen again.
		signal(signo, SIG_DFL);
	}
}

// Set up UNIX signal handlers to catch signals caused by VX code
int vxproc_siginit(void)
{
	// See if there's already an alternate signal stack
	stack_t ss;
	if (sigaltstack(NULL, &ss) < 0)
		return -1;
	assert(!(ss.ss_flags & SS_ONSTACK));
	if (ss.ss_flags & SS_DISABLE) {

		// Allocate an alternate signal stack.
		void *stk = malloc(SIGSTKSZ);
		if (stk == NULL)
			return -1;
	//	fprintf(stderr, "signal stack: %016lx-%016lx (%dKB)\n",
	//		stk, stk+SIGSTKSZ, SIGSTKSZ/1024);

		// Register it
		ss.ss_flags = 0;
		ss.ss_sp = stk;
		ss.ss_size = SIGSTKSZ;
		if (sigaltstack(&ss, NULL) < 0) {
			free(stk);
			return -1;
		}
	}

	// Register our signal handler
	struct sigaction sa;
	sa.sa_handler = (void*)sighandler;
	sa.sa_flags = SA_ONSTACK | SA_SIGINFO;
	if (sigemptyset(&sa.sa_mask) < 0)
		return -1;
	if (sigaction(SIGSEGV, &sa, NULL) < 0)
		return -1;
	if (sigaction(SIGBUS, &sa, NULL) < 0)
		return -1;
	if (sigaction(SIGFPE, &sa, NULL) < 0)
		return -1;

	return 0;
}


void vxproc_free(vxproc *proc)
{
	assert(procs[proc->vxpno] == proc);
	procs[proc->vxpno] = NULL;

	// Free the emulation state
	if (proc->emu != NULL)
		vxemu_free(proc->emu);

	// Unmap the process's data segment
	if (proc->data != NULL)
		munmap(proc->data, proc->size);

	// Release the file descriptor holding the process's data
	if (proc->fd >= 0)
		close(proc->fd);

	free(proc);
}

