
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "vx/proc.h"
#include "vx/vx32.h"

#include "env.h"
#include "emu.h"
#include "elf.h"
#include "words.h"


// Helper function to copy the argument and environment arrays on load
static uint32_t copyargs(vxproc *proc, uint32_t *espp, uint32_t stackhi,
			const char *const argp [],
			uint32_t *argcp)
{
	uint32_t esp = *espp;

	// Gracefully handle a null array pointer
	static const char * const nullp = NULL;
	if (argp == NULL)
		argp = &nullp;

	// Count the number of args in this array
	uint32_t argc = 0;
	while (argp[argc] != NULL)
		argc++;
	if (argc > VX32_ARG_MAX/4)
		argc = VX32_ARG_MAX/4;

	// Allocate space for the array, making sure it's word-aligned
	esp -= (argc + 1) * 4;
	esp &= ~3;
	uint32_t vargva = esp;
	uint32_t *vargp = proc->data + vargva;

	// Copy the arguments
	for (unsigned i = 0; i < argc; i++) {

		// Make sure we're not out of argument space
		if ((stackhi - esp) >= VX32_ARG_MAX)
			break;

		// Check length of this string
		const char *arg = argp[i];
		unsigned len = strlen(arg);
		if (len > VX32_ARG_MAX)
			len = VX32_ARG_MAX;

		// Copy it
		esp -= len+1;
		memcpy(proc->data + esp, arg, len+1);

		// Stash a pointer into the varg array
		vargp[i] = htol32(esp);
	}
	vargp[argc] = 0;

	// Update the stack pointer (not bothering about alignment for now)
	*espp = esp;

	// Optionally return the length of the array
	if (argcp)
		*argcp = argc;

	return vargva;
}

int vx32_load(vxproc *proc,
		ssize_t (*readcb)(void *cbdata, off_t ofs,
				void *buf, size_t size),
		void *cbdata,
		const char *const argv [], const char *const envp [])
{
	// Clear the process's address space
	vxproc_clear(proc);

	// Read and check the ELF header
	struct Elf32 h;
	ssize_t act = readcb(cbdata, 0, &h, sizeof(h));
	if (act < 0)
		return -1;
	if (act < sizeof(h)) {
		noexec:
		errno = ENOEXEC;
		return -1;
	}
	if (ltoh32(h.e_magic) != ELF_MAGIC)
		goto noexec;

	// Read the program header table
	off_t phoff = ltoh32(h.e_phoff);
	size_t phnum = ltoh16(h.e_phnum);
	if (phnum <= 0 || phnum > 32)	// arbitrary limit for security
		goto noexec;
	struct Proghdr ph[phnum];
	act = readcb(cbdata, phoff, ph, sizeof(ph));
	if (act < 0)
		return -1;
	if (act < sizeof(ph))
		goto noexec;

	// Load each program segment
	size_t stackhi = VXMEMSIZE;
	for (size_t i = 0; i < phnum; i++) {
		const struct Proghdr *p = &ph[i];
		if (ltoh32(p->p_type) != ELF_PROG_LOAD)
			continue;

		// Validate the program segment
		off_t offset = ltoh32(p->p_offset);
		size_t filesz = ltoh32(p->p_filesz);
		size_t va = ltoh32(p->p_va);
		size_t memsz = ltoh32(p->p_memsz);
		if (filesz > memsz)
			goto noexec;

		// Validate the memory page region the segment loads into
		size_t memlo = VXPAGETRUNC(va);
		size_t memhi = VXPAGEROUND(va + memsz);
		if (memlo > VXMEMSIZE || memhi > VXMEMSIZE || memhi < memlo)
			goto noexec;

		// Make sure the VX process is big enough, and mapped
		if (proc->size < memhi)
			if (vxproc_resize(proc, memhi) < 0)
				return -1;
		if (vxproc_map(proc) < 0)
			return -1;

		// Temporarily give ourselves write permissions
		// on the segment in order to load it.
		if (vxproc_setperm(proc, memlo, memhi - memlo,
				VXPERM_READ | VXPERM_WRITE) < 0)
			return -1;

		// Load the segment.
		// Any bss portion is already cleared automatically
		// by virtue of the vxproc_clear() above.
		act = readcb(cbdata, offset, proc->data + va, filesz);
		if (act < 0)
			return -1;
		if (act < filesz)
			goto noexec;

		// Set permissions appropriately on the segment's pages
		int flags = ltoh32(p->p_flags);
		int perm = 0;
		switch (flags & (ELF_PROG_FLAG_READ | ELF_PROG_FLAG_WRITE |
				ELF_PROG_FLAG_EXEC)) {
		case ELF_PROG_FLAG_READ:
			perm = VXPERM_READ;
			break;
		case ELF_PROG_FLAG_READ | ELF_PROG_FLAG_WRITE:
			perm = VXPERM_READ | VXPERM_WRITE;
			break;
		case ELF_PROG_FLAG_READ | ELF_PROG_FLAG_EXEC:
			perm = VXPERM_READ | VXPERM_EXEC;
			break;
		default:
			goto noexec;	// invalid perms
		}
		if (vxproc_setperm(proc, memlo, memhi - memlo, perm) < 0)
			return -1;

		// Find the lowest-used va, for locating the stack
		if (va < stackhi)
			stackhi = va;
	}

	// Set up the process's stack,
	// growing downward from the executable's base load address.
	// Initially we enable read/write access on 64K of stack,
	// but the process is free to change that if it wants a bigger stack.
	stackhi = VXPAGETRUNC(stackhi);
	if (stackhi < VX32_STACK)
		goto noexec;
	if (vxproc_setperm(proc, stackhi - VX32_STACK, VX32_STACK,
			VXPERM_READ | VXPERM_WRITE) < 0)
		return -1;
	proc->emu->reg[ESP] = stackhi;

	// Push the argument and environment arrays on the stack.
	uint32_t esp = stackhi;
	uint32_t argc;
	uint32_t vargp = copyargs(proc, &esp, stackhi, argv, &argc);
	uint32_t venvp = copyargs(proc, &esp, stackhi, envp, NULL);
	if ((stackhi - esp) >= VX32_ARG_MAX) {
		errno = E2BIG;
		return -1;
	}

	// Push the three actual arguments to main(),
	// making sure we end up aligned on a 16-byte boundary.
	esp -= 3*4;	// argc, argv, envp
	esp &= ~15;	// align
	uint32_t *stackp = proc->data + esp;
	stackp[0] = htol32(argc);
	stackp[1] = htol32(vargp);
	stackp[2] = htol32(venvp);

	// Set up the process's initial register state
	for (int i = 0; i < 8; i++)
		proc->emu->reg[i] = 0;
	proc->emu->reg[ESP] = esp;
	proc->emu->eflags = 0;
	proc->emu->eip = ltoh32(h.e_entry);

	return 0;
}

void vx32_getregs(struct vxproc *proc, struct vx32regs *regs)
{
	vxemu *emu = proc->emu;

	regs->eax = emu->reg[EAX];
	regs->ecx = emu->reg[ECX];
	regs->edx = emu->reg[EDX];
	regs->ebx = emu->reg[EBX];
	regs->esp = emu->reg[ESP];
	regs->ebp = emu->reg[EBP];
	regs->esi = emu->reg[ESI];
	regs->edi = emu->reg[EDI];

	regs->eip = emu->eip;
	regs->eflags = emu->eflags;
}

void vx32_setregs(struct vxproc *proc, const struct vx32regs *regs)
{
	vxemu *emu = proc->emu;

	emu->reg[EAX] = regs->eax;
	emu->reg[ECX] = regs->ecx;
	emu->reg[EDX] = regs->edx;
	emu->reg[EBX] = regs->ebx;
	emu->reg[ESP] = regs->esp;
	emu->reg[EBP] = regs->ebp;
	emu->reg[ESI] = regs->esi;
	emu->reg[EDI] = regs->edi;

	emu->eip = regs->eip;
	emu->eflags = regs->eflags & VX32_EFLAGS;
}

