/*
 * Portable (but slow) vx32 instruction set emulation.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <setjmp.h>
#include <assert.h>
#include <errno.h>
#include <sys/mman.h>

#include "env/env.h"
#include "emu.h"
#include "words.h"



// Parity lookup table for calculating the PF bit in EFLAGS
uint8_t paritytab[256] = {
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
};



// Create the emulation state for a new process
int vxemu_init(struct vxproc *vxp)
{
	vxemu *e = calloc(1, sizeof(*e));
	if (e == NULL)
		return -1;

	vxp->emu = e;
	e->proc = vxp;

	return 0;
}

void vxemu_free(vxemu *e)
{
	assert(e->proc->emu == e);
	e->proc->emu = NULL;

	// Free the vxemu state
	free(e);
}






static inline int checkperm(vxemu *e, int32_t va)
{
	if (va >= e->memsize)
		return 0;
	return e->proc->perm[va >> VXPAGESHIFT];
}


// Update the values of all the status flags in the EFLAGS register,
// based on the last ALU operation done according to e->stmode.
static void update_eflags(vxemu *e)
{
	unsigned efl = 0;
	int32_t result;

	// Compute the last ALU instruction's numeric result
	// and the values of the carry and overflow flags.
	switch (e->stmode) {

	case STM_EFLAGS:	// Already up-to-date
		return;

	// 8-bit operand size
	case STM_8 | STM_LOG:
		result = (int8_t)e->starg1;
		break;

	case STM_8 | STM_ADD: {
		int32_t result32 = (int8_t)e->starg1 + (int8_t)e->starg2;
		result = (int8_t)result32;
		if ((uint8_t)result32 < (uint8_t)e->starg1)
			efl |= EFL_CF;
		if (result32 != result)
			efl |= EFL_OF;
		break;
	    }

	case STM_8 | STM_ADDC: {
		int32_t result32 = (int8_t)e->starg1 + (int8_t)e->starg2 + 1;
		result = (int8_t)result32;
		if ((uint8_t)result32 <= (uint8_t)e->starg1)
			efl |= EFL_CF;
		if (result32 != result)
			efl |= EFL_OF;
		break;
	    }

	case STM_8 | STM_SUB: {
		int32_t result32 = (int8_t)e->starg1 - (int8_t)e->starg2;
		result = (int8_t)result32;
		if ((uint8_t)result32 > (uint8_t)e->starg1)
			efl |= EFL_CF;
		if (result32 != result)
			efl |= EFL_OF;
		break;
	    }

	case STM_8 | STM_SUBB: {
		int32_t result32 = (int8_t)e->starg1 - (int8_t)e->starg2 - 1;
		result = (int8_t)result32;
		if ((uint8_t)result32 >= (uint8_t)e->starg1)
			efl |= EFL_CF;
		if (result32 != result)
			efl |= EFL_OF;
		break;
	    }

	case STM_8 | STM_MUL: {
		uint32_t result32 = (uint8_t)e->starg1 * (uint8_t)e->starg2;
		result = (int8_t)result32;
		if (result32 != (uint8_t)result32)
			efl |= EFL_CF | EFL_OF;
		break;
	    }

	case STM_8 | STM_IMUL: {
		int32_t result32 = (int8_t)e->starg1 * (int8_t)e->starg2;
		result = (int8_t)result32;
		if (result32 != result)
			efl |= EFL_CF | EFL_OF;
		break;
	    }

#if 0	// XXX
	case STM_8 | STM_ROL:
	case STM_8 | STM_RCL:
		assert(0); // XXX can't differentiate between cnt=0 and cnt>=8
	case STM_8 | STM_SHL:
		return (((e->starg1 & 0xff) << e->starg2) >> 8);
	case STM_8 | STM_RCLC:
		return (((e->starg1 | 0x100) << e->starg2) >> 8);
	case STM_8 | STM_ROR:
	case STM_8 | STM_RCR:
		assert(0); // XXX can't differentiate between cnt=0 and cnt>=8
	case STM_8 | STM_SHR:
		return (((e->starg1 & 0xff) << 1) >> e->starg2) & 1;
	case STM_8 | STM_RCRC:
		return (((e->starg1 << 1) | 1) >> e->starg2) & 1;
	case STM_8 | STM_SAR:
		return ((((int32_t)(int8_t)e->starg1) << 1) >> e->starg2) & 1;
#endif

	// 16-bit operand size
	case STM_16 | STM_LOG:
		result = (int16_t)e->starg1;
		break;

	case STM_16 | STM_ADD: {
		int32_t result32 = (int16_t)e->starg1 + (int16_t)e->starg2;
		result = (int16_t)result32;
		if ((uint16_t)result32 < (uint16_t)e->starg1)
			efl |= EFL_CF;
		if (result32 != result)
			efl |= EFL_OF;
		break;
	    }

	case STM_16 | STM_ADDC: {
		int32_t result32 = (int16_t)e->starg1 + (int16_t)e->starg2 + 1;
		result = (int16_t)result32;
		if ((uint16_t)result32 <= (uint16_t)e->starg1)
			efl |= EFL_CF;
		if (result32 != result)
			efl |= EFL_OF;
		break;
	    }

	case STM_16 | STM_SUB: {
		int32_t result32 = (int16_t)e->starg1 - (int16_t)e->starg2;
		result = (int16_t)result32;
		if ((uint16_t)result32 > (uint16_t)e->starg1)
			efl |= EFL_CF;
		if (result32 != result)
			efl |= EFL_OF;
		break;
	    }

	case STM_16 | STM_SUBB: {
		int32_t result32 = (int16_t)e->starg1 - (int16_t)e->starg2 - 1;
		result = (int16_t)result32;
		if ((uint16_t)result32 >= (uint16_t)e->starg1)
			efl |= EFL_CF;
		if (result32 != result)
			efl |= EFL_OF;
		break;
	    }

	case STM_16 | STM_MUL: {
		uint32_t result32 = (uint16_t)e->starg1 * (uint16_t)e->starg2;
		result = (int16_t)result32;
		if (result32 != (uint16_t)result32)
			efl |= EFL_CF | EFL_OF;
		break;
	    }

	case STM_16 | STM_IMUL: {
		int32_t result32 = (int16_t)e->starg1 * (int16_t)e->starg2;
		result = (int16_t)result32;
		if (result32 != result)
			efl |= EFL_CF | EFL_OF;
		break;
	    }

#if 0	// XXX
	case STM_16 | STM_ROL:
	case STM_16 | STM_RCL:
		assert(0); // XXX can't differentiate between cnt=0 and cnt>=8
	case STM_16 | STM_SHL:
		return (((e->starg1 & 0xffff) << e->starg2) >> 16);
	case STM_16 | STM_RCLC:
		return (((e->starg1 | 0x10000) << e->starg2) >> 16);
	case STM_16 | STM_ROR:
	case STM_16 | STM_RCR:
		assert(0); // XXX can't differentiate between cnt=0 and cnt>=8
	case STM_16 | STM_SHR:
		return (((e->starg1 & 0xffff) << 1) >> e->starg2) & 1;
	case STM_16 | STM_RCRC:
		return (((e->starg1 << 1) | 1) >> e->starg2) & 1;
	case STM_16 | STM_SAR:
		return ((((int32_t)(int16_t)e->starg1) << 1) >> e->starg2) & 1;
#endif

	// 32-bit operand size
	case STM_32 | STM_LOG:
		result = (int32_t)e->starg1;
		break;

	case STM_32 | STM_ADD: {
		int64_t result64 = (int64_t)(int32_t)e->starg1 +
					(int64_t)(int32_t)e->starg2;
		result = (int32_t)result64;
		if ((uint32_t)result64 < (uint32_t)e->starg1)
			efl |= EFL_CF;
		if (result64 != result)
			efl |= EFL_OF;
		break;
	    }

	case STM_32 | STM_ADDC: {
		int64_t result64 = (int64_t)(int32_t)e->starg1 +
					(int64_t)(int32_t)e->starg2 + 1;
		result = (int32_t)result64;
		if ((uint32_t)result64 <= (uint32_t)e->starg1)
			efl |= EFL_CF;
		if (result64 != result)
			efl |= EFL_OF;
		break;
	    }

	case STM_32 | STM_SUB: {
		int64_t result64 = (int64_t)(int32_t)e->starg1 -
					(int64_t)(int32_t)e->starg2;
		result = (int32_t)result64;
		if ((uint32_t)result64 > (uint32_t)e->starg1)
			efl |= EFL_CF;
		if (result64 != result)
			efl |= EFL_OF;
		break;
	    }

	case STM_32 | STM_SUBB: {
		int64_t result64 = (int64_t)(int32_t)e->starg1 -
					(int64_t)(int32_t)e->starg2 - 1;
		result = (int32_t)result64;
		if ((uint32_t)result64 >= (uint32_t)e->starg1)
			efl |= EFL_CF;
		if (result64 != result)
			efl |= EFL_OF;
		break;
	    }

	case STM_32 | STM_MUL: {
		uint64_t result64 = (uint64_t)(uint32_t)e->starg1 *
					(uint64_t)(uint32_t)e->starg2;
		result = (int32_t)result64;
		if (result64 != (uint32_t)result64)
			efl |= EFL_CF | EFL_OF;
		break;
	    }

	case STM_32 | STM_IMUL: {
		int64_t result64 = (int64_t)(int32_t)e->starg1 *
					(int64_t)(int32_t)e->starg2;
		result = (int32_t)result64;
		if (result64 != result)
			efl |= EFL_CF | EFL_OF;
		break;
	    }

#if 0	// XXX
	case STM_32 | STM_ROL:
	case STM_32 | STM_RCL:
	case STM_32 | STM_RCLC:
	case STM_32 | STM_SHL:
		return e->starg2 ? ((e->starg1 >> (32 - e->starg2)) & 1) : 0;
	case STM_32 | STM_ROR:
	case STM_32 | STM_RCR:
	case STM_32 | STM_RCRC:
	case STM_32 | STM_SHR:
	case STM_32 | STM_SAR:
		return e->starg2 ? ((e->starg1 >> (e->starg2 - 1)) & 1) : 0;
#endif

	default:
		assert(0);
	}

	// Compute the generic bits that only depend on the result.
	if (result < 0)
		efl |= EFL_SF;
	if (result == 0)
		efl |= EFL_ZF;
	if (paritytab[(uint8_t)result])
		efl |= EFL_PF;

	e->eflags = efl;
	e->stmode = STM_EFLAGS;
}

// Test the current value of particular EFLAGS bits,
// based on the last ALU operation done according to e->stmode.
// These functions fall back to update_eflags() for less common cases.
static unsigned test_cf(vxemu *e)
{
	switch (e->stmode) {
	case STM_EFLAGS:
		return (e->eflags & EFL_CF) != 0;

	// 8-bit operand size
	case STM_8 | STM_LOG:
		return 0;	// CF is always 0 after logical instructions
	case STM_8 | STM_ADD:
		return (uint8_t)(e->starg1+e->starg2) < (uint8_t)e->starg1;
	case STM_8 | STM_ADDC:
		return (uint8_t)(e->starg1+e->starg2+1) <= (uint8_t)e->starg1;
	case STM_8 | STM_SUB:
		return ((uint8_t)e->starg1 < (uint8_t)e->starg2);
	case STM_8 | STM_SUBB:
		return ((uint8_t)e->starg1 <= (uint8_t)e->starg2);

	// 16-bit operand size
	case STM_16 | STM_LOG:
		return 0;	// CF is always 0 after logical instructions
	case STM_16 | STM_ADD:
		return (uint16_t)(e->starg1+e->starg2) < (uint16_t)e->starg1;
	case STM_16 | STM_ADDC:
		return (uint16_t)(e->starg1+e->starg2+1) <= (uint16_t)e->starg1;
	case STM_16 | STM_SUB:
		return ((uint16_t)e->starg1 < (uint16_t)e->starg2);
	case STM_16 | STM_SUBB:
		return ((uint16_t)e->starg1 <= (uint16_t)e->starg2);

	// 32-bit operand size
	case STM_32 | STM_LOG:
		return 0;	// CF is always 0 after logical instructions
	case STM_32 | STM_ADD:
		return (uint32_t)(e->starg1+e->starg2) < (uint32_t)e->starg1;
	case STM_32 | STM_ADDC:
		return (uint32_t)(e->starg1+e->starg2+1) <= (uint32_t)e->starg1;
	case STM_32 | STM_SUB:
		return ((uint32_t)e->starg1 < (uint32_t)e->starg2);
	case STM_32 | STM_SUBB:
		return ((uint32_t)e->starg1 <= (uint32_t)e->starg2);

	// Handle other results by updating all status bits at once.
	default:
		update_eflags(e);
		return (e->eflags & EFL_CF) != 0;
	}
}

static unsigned test_sf(vxemu *e)
{
	switch (e->stmode) {
	case STM_EFLAGS:
		return (e->eflags & EFL_SF) != 0;

	// 8-bit operand size
	case STM_8 | STM_LOG:
		return ((int8_t)e->starg1 < 0);		// starg1 is result

	// 16-bit operand size
	case STM_16 | STM_LOG:
		return ((int16_t)e->starg1 < 0);	// starg1 is result

	// 32-bit operand size
	case STM_32 | STM_LOG:
		return ((int32_t)e->starg1 < 0);	// starg1 is result

	// Handle other results by updating all status bits at once.
	default:
		update_eflags(e);
		return (e->eflags & EFL_SF) != 0;
	}
}

static unsigned test_zf(vxemu *e)
{
	switch (e->stmode) {
	case STM_EFLAGS:
		return (e->eflags & EFL_ZF) != 0;

	// 8-bit operand size
	case STM_8 | STM_LOG:
		return (uint8_t)e->starg1 == 0;		// starg1 is result
	case STM_8 | STM_ADD:
		return (uint8_t)(e->starg1+e->starg2) == 0;
	case STM_8 | STM_ADDC:
		return (uint8_t)(e->starg1+e->starg2+1) == 0;
	case STM_8 | STM_SUB:
		return (uint8_t)(e->starg1-e->starg2) == 0;
	case STM_8 | STM_SUBB:
		return (uint8_t)(e->starg1-e->starg2-1) == 0;

	// 16-bit operand size
	case STM_16 | STM_LOG:
		return (uint16_t)e->starg1 == 0;	// starg1 is result
	case STM_16 | STM_ADD:
		return (uint16_t)(e->starg1+e->starg2) == 0;
	case STM_16 | STM_ADDC:
		return (uint16_t)(e->starg1+e->starg2+1) == 0;
	case STM_16 | STM_SUB:
		return (uint16_t)(e->starg1-e->starg2) == 0;
	case STM_16 | STM_SUBB:
		return (uint16_t)(e->starg1-e->starg2-1) == 0;

	// 32-bit operand size
	case STM_32 | STM_LOG:
		return (uint32_t)e->starg1 == 0;	// starg1 is result
	case STM_32 | STM_ADD:
		return (uint32_t)(e->starg1+e->starg2) == 0;
	case STM_32 | STM_ADDC:
		return (uint32_t)(e->starg1+e->starg2+1) == 0;
	case STM_32 | STM_SUB:
		return (uint32_t)(e->starg1-e->starg2) == 0;
	case STM_32 | STM_SUBB:
		return (uint32_t)(e->starg1-e->starg2-1) == 0;

	// Handle other results by updating all status bits at once.
	default:
		update_eflags(e);
		return (e->eflags & EFL_ZF) != 0;
	}
}

static unsigned test_of(vxemu *e)
{
	// Don't bother optimizing; just update all status flags at once.
	update_eflags(e);
	return (e->eflags & EFL_OF) != 0;
}

static unsigned test_pf(vxemu *e)
{
	// Don't bother optimizing; just update all status flags at once.
	update_eflags(e);
	return (e->eflags & EFL_PF) != 0;
}

// Test unsigned Below or Equal condition (CF=1 or ZF=1),
// optimizing the common cases of CMP and SUB instructions.
static unsigned test_be(vxemu *e)
{
	switch (e->stmode) {

	// Handle common comparison cases.
	case STM_8 | STM_SUB:
		return ((uint8_t)e->starg1 <= (uint8_t)e->starg2);
	case STM_16 | STM_SUB:
		return ((uint16_t)e->starg1 <= (uint16_t)e->starg2);
	case STM_32 | STM_SUB:
		return ((uint32_t)e->starg1 <= (uint32_t)e->starg2);

	// Handle other results by updating all status bits at once.
	default:
		update_eflags(e);
	case STM_EFLAGS:
		return (e->eflags & (EFL_CF | EFL_ZF)) != 0;
	}
}

// Test signed Less than condition (SF != OF),
// optimizing the common cases of CMP and SUB instructions.
static unsigned test_l(vxemu *e)
{
	switch (e->stmode) {

	// Handle common comparison cases.
	case STM_8 | STM_SUB:
		return ((int8_t)e->starg1 < (int8_t)e->starg2);
	case STM_16 | STM_SUB:
		return ((int16_t)e->starg1 < (int16_t)e->starg2);
	case STM_32 | STM_SUB:
		return ((int32_t)e->starg1 < (int32_t)e->starg2);

	// Handle other results by updating all status bits at once.
	default:
		update_eflags(e);
	case STM_EFLAGS:
		return ((e->eflags >> EFL_SF_BIT) ^
			(e->eflags >> EFL_OF_BIT)) & 1;
	}
}

// Test signed Less or Equal condition (ZF = 1 or SF != OF),
// optimizing the common cases of CMP and SUB instructions.
static unsigned test_le(vxemu *e)
{
	switch (e->stmode) {

	// Handle common comparison cases.
	case STM_8 | STM_SUB:
		return ((int8_t)e->starg1 <= (int8_t)e->starg2);
	case STM_16 | STM_SUB:
		return ((int16_t)e->starg1 <= (int16_t)e->starg2);
	case STM_32 | STM_SUB:
		return ((int32_t)e->starg1 <= (int32_t)e->starg2);

	// Handle other results by updating all status bits at once.
	default:
		update_eflags(e);
	case STM_EFLAGS:
		return ((e->eflags >> EFL_ZF_BIT) |
			((e->eflags >> EFL_SF_BIT) ^
			 (e->eflags >> EFL_OF_BIT))) & 1;
	}
}



// Macros to extract fields in a Mod-Reg-R/M byte
#define EA_MOD(b)	((uint8_t)(b) >> 6)
#define EA_REG(b)	(((uint8_t)(b) >> 3) & 7)
#define EA_RM(b)	((uint8_t)(b) & 7)
#define EA_ISREG(b)	(EA_MOD(b) == 3)
#define EA_ISMEM(b)	(EA_MOD(b) != 3)

#define SIB_SCALE(b)	((uint8_t)(b) >> 6)
#define SIB_INDEX(b)	(((uint8_t)(b) >> 3) & 7)
#define SIB_BASE(b)	((uint8_t)(b) & 7)


#define PUTREG8(r, val) \
	if ((r) < 4) /* AL/CL/DL/BL */ \
		e->reg[r] = (e->reg[r] & 0xffffff00) | ((val) & 0xff); \
	else /* AH/CH/DH/BH */ \
		e->reg[(r) & 3] = (e->reg[r] & 0xffff00ff) | \
				(((val) & 0xff) << 8);

#define PUTREG16(r, val) \
	e->reg[r] = (e->reg[r] & 0xffff0000) | ((val) & 0xffff);

#define PUTREG32(r, val) \
	e->reg[r] = (val);


// Scan a Mod-Reg-R/M byte and compute its memory effective address.
// Does NOT check if the effective address is in range.
// Must only be called if EA_ISMEM(eabyte).
static uint32_t scanmea(vxemu *e, unsigned eabyte, uint8_t **inpp)
{
	register uint8_t *inp = *inpp;	// initial instruction decode pointer
	register uint32_t addr;		// effective address to compute

	unsigned rm = EA_RM(eabyte);
	switch (EA_MOD(eabyte)) {
	case 0:
		switch (rm) {
		case 4:	{ // SIB
			unsigned sib = *inp++;
			unsigned indexreg = SIB_INDEX(sib);
			unsigned basereg = SIB_BASE(sib);
			if (basereg == 5) {	// special disp32 mode, no base
				addr = getu32l(inp);
				inp += 4;
			} else {		// regular base register
				addr = e->reg[basereg];
			}
			if (indexreg != 4) {	// handle index if present
				unsigned index = e->reg[indexreg];
				switch (SIB_SCALE(sib)) {
				case 1: index *= 2; break;
				case 2: index *= 4; break;
				case 3: index *= 8; break;
				}
				addr += index;
			}
			break;
		    }
		case 5:	// disp32
			addr = getu32l(inp); inp += 4;
			break;
		default: // [reg], no displacement
			addr = e->reg[rm];
			break;
		}
		break;

	case 1:
		switch (rm) {
		case 4:	{ // SIB+disp8
			unsigned sib = inp[0];
			addr = (int32_t)(int8_t)inp[1];
			inp += 2;
			addr += e->reg[SIB_BASE(sib)];
			unsigned indexreg = SIB_INDEX(sib);
			if (indexreg != 4) {	// handle index if present
				unsigned index = e->reg[indexreg];
				switch (SIB_SCALE(sib)) {
				case 1: index *= 2; break;
				case 2: index *= 4; break;
				case 3: index *= 8; break;
				}
				addr += index;
			}
			break;
		    }
		default: { // [reg]+disp8
			int8_t disp8 = *inp++;	// NB: signed
			addr = e->reg[rm] + (int32_t)disp8;
			break;
		    }
		}
		break;

	case 2:
		switch (rm) {
		case 4: { // SIB+disp32
			unsigned sib = inp[0];
			addr = getu32l(inp+1);
			inp += 1+4;
			addr += e->reg[SIB_BASE(sib)];
			unsigned indexreg = SIB_INDEX(sib);
			if (indexreg != 4) {	// handle index if present
				unsigned index = e->reg[indexreg];
				switch (SIB_SCALE(sib)) {
				case 1: index *= 2; break;
				case 2: index *= 4; break;
				case 3: index *= 8; break;
				}
				addr += index;
			}
			break;
		    }
		default: { // [reg]+disp32
			uint32_t disp32 = getu32l(inp); inp += 4;
			addr = e->reg[rm] + disp32;
			break;
		    }
		}
		break;

	case 3:
		assert(0);
	}

	*inpp = inp;
	return addr;
}

#if 0
// Scan a Mod-Reg-R/M byte and compute its effective address.
// If it's a register effective address,
// computes a pointer to the appropriate portion of the e->reg[] array.
// If it's a memory effective address, ensures it is in range.
static uint8_t *scan_ea8(vxemu *e, unsigned eabyte, uint8_t *inpp, int check)
{
	if (EA_ISREG(eabyte)) {
		unsigned r = EA_RM(eabyte);
#ifdef WORDS_BIGENDIAN
		return (uint8_t*)&e->reg[(eabyte >> 3) & 3] +
				((eabyte & 0x20) ? 3 : 2);
#else
		return (uint8_t*)&e->reg[(eabyte >> 3) & 3] +
				((eabyte & 0x20) ? 1 : 0);
#endif
	} else {
		uint32_t mea = scan_mea(e, eabyte, inpp, 
	}
}
#endif


#define CHECKMEA(e, eabyte, inp, meavar) \
	uint32_t meavar = scanmea(e, eabyte, &inp); \
	if (meavar >= e->memsize) assert(0); /* XXX */


#define LOADRM8(e,eabyte,inp,result) \
	register uint8_t result; \
	if (EA_ISREG(eabyte)) { \
		result = e->reg[EA_RM(eabyte)]; \
	} else { \
		CHECKMEA(e, eabyte, inp, mea); \
		result = e->mem[mea]; \
	}

#define LOADRM16(e,eabyte,inp,result) \
	register uint16_t result; \
	if (EA_ISREG(eabyte)) { \
		result = e->reg[EA_RM(eabyte)]; \
	} else { \
		CHECKMEA(e, eabyte, inp, mea); \
		result = getu16l(e->mem + mea); \
	}

#define LOADRM32(e,eabyte,inp,result) \
	register uint32_t result; \
	if (EA_ISREG(eabyte)) { \
		result = e->reg[EA_RM(eabyte)]; \
	} else { \
		CHECKMEA(e, eabyte, inp, mea); \
		result = getu32l(e->mem + mea); \
	}


#define SETST(e,mode,dstarg,srcarg,result) \
	if ((mode & STM_MODE) != 0) { \
		e->stmode = mode; \
		if ((mode & STM_MODE) != STM_LOG) { \
			e->starg1 = dstarg; \
			e->starg2 = srcarg; \
		} else \
			e->starg1 = result; \
	}


// After fully decoding any instruction,
// we must check the final inp against inmax
// to prevent executing into a readable but not executable page.
// Each instruction must perform this check BEFORE committing any results.
#define CHECKINP	if (inp >= inmax) goto noexec;


// ALU instruction, reg dest, r/m source.
#define ALU_REGRM8(e, expr, stmode, writeback) { \
		register uint8_t eabyte = *inp++; \
		register unsigned dstreg = EA_REG(eabyte); \
		register uint32_t dstarg = e->reg[dstreg]; \
		LOADRM8(e, eabyte, inp, srcarg); \
		CHECKINP; \
		register uint32_t result = (expr); \
		if (writeback) PUTREG8(dstreg, result); \
		SETST(e, stmode | STM_8, dstarg, srcarg, result); \
	}

#define ALU_REGRM16(e, expr, stmode, writeback) { \
		register uint8_t eabyte = *inp++; \
		register unsigned dstreg = EA_REG(eabyte); \
		register uint32_t dstarg = e->reg[dstreg]; \
		LOADRM16(e, eabyte, inp, srcarg); \
		CHECKINP; \
		register uint32_t result = (expr); \
		if (writeback) PUTREG16(dstreg, result); \
		SETST(e, stmode | STM_16, dstarg, srcarg, result); \
	}

#define ALU_REGRM32(e, expr, stmode, writeback) { \
		register uint8_t eabyte = *inp++; \
		register unsigned dstreg = EA_REG(eabyte); \
		register uint32_t dstarg = e->reg[dstreg]; \
		LOADRM32(e, eabyte, inp, srcarg); \
		CHECKINP; \
		register uint32_t result = (expr); \
		if (writeback) PUTREG32(dstreg, result); \
		SETST(e, stmode | STM_32, dstarg, srcarg, result); \
	}


// ALU instruction, r/m dest, reg source.
#define ALU_RMREG8(e, expr, stmode, writeback) { \
		register uint8_t eabyte = *inp++; \
		register uint32_t srcarg = e->reg[EA_REG(eabyte)]; \
		register uint32_t dstarg, result; \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			CHECKINP; \
			dstarg = e->reg[dstreg]; \
			result = (expr); \
			if (writeback) PUTREG8(dstreg, result); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			CHECKINP; \
			dstarg = e->mem[mea]; \
			result = (expr); \
			if (writeback) e->mem[mea] = result; \
		} \
		SETST(e, stmode | STM_8, dstarg, srcarg, result); \
	}

#define ALU_RMREG16(e, expr, stmode, writeback) { \
		register uint8_t eabyte = *inp++; \
		register uint32_t srcarg = e->reg[EA_REG(eabyte)]; \
		register uint32_t dstarg, result; \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			CHECKINP; \
			dstarg = e->reg[dstreg]; \
			result = (expr); \
			if (writeback) PUTREG16(dstreg, result); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			CHECKINP; \
			dstarg = getu16l(e->mem + mea); \
			result = (expr); \
			if (writeback) putu16l(e->mem + mea, result); \
		} \
		SETST(e, stmode | STM_16, dstarg, srcarg, result); \
	}

#define ALU_RMREG32(e, expr, stmode, writeback) { \
		register uint8_t eabyte = *inp++; \
		register uint32_t srcarg = e->reg[EA_REG(eabyte)]; \
		register uint32_t dstarg, result; \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			CHECKINP; \
			dstarg = e->reg[dstreg]; \
			result = (expr); \
			if (writeback) PUTREG32(dstreg, result); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			CHECKINP; \
			dstarg = getu32l(e->mem + mea); \
			result = (expr); \
			if (writeback) putu32l(e->mem + mea, result); \
		} \
		SETST(e, stmode | STM_32, dstarg, srcarg, result); \
	}


// ALU instruction, r/m dest, immediate source.
// Assumes eabyte already scanned.
#define ALU_RMIMM8(e, expr, stmode, writeback) { \
		register uint32_t dstarg, srcarg, result; \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			srcarg = *inp++; \
			dstarg = e->reg[dstreg]; \
			CHECKINP; \
			result = (expr); \
			if (writeback) PUTREG8(dstreg, result); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			srcarg = *inp++; \
			dstarg = e->mem[mea]; \
			CHECKINP; \
			result = (expr); \
			if (writeback) e->mem[mea] = result; \
		} \
		SETST(e, stmode | STM_8, dstarg, srcarg, result); \
	}

#define ALU_RMIMM16(e, expr, stmode, writeback) { \
		register uint32_t dstarg, srcarg, result; \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			srcarg = getu16l(inp); inp += 2; \
			dstarg = e->reg[dstreg]; \
			CHECKINP; \
			result = (expr); \
			if (writeback) PUTREG16(dstreg, result); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			srcarg = getu16l(inp); inp += 2; \
			dstarg = getu16l(e->mem + mea); \
			CHECKINP; \
			result = (expr); \
			if (writeback) putu16l(e->mem + mea, result); \
		} \
		SETST(e, stmode | STM_16, dstarg, srcarg, result); \
	}

#define ALU_RMIMM32(e, expr, stmode, writeback) { \
		register uint32_t dstarg, srcarg, result; \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			srcarg = getu32l(inp); inp += 4; \
			dstarg = e->reg[dstreg]; \
			CHECKINP; \
			result = (expr); \
			if (writeback) PUTREG32(dstreg, result); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			srcarg = getu32l(inp); inp += 4; \
			dstarg = getu32l(e->mem + mea); \
			CHECKINP; \
			result = (expr); \
			if (writeback) putu32l(e->mem + mea, result); \
		} \
		SETST(e, stmode | STM_32, dstarg, srcarg, result); \
	}


// ALU instruction, r/m dest, sign-extended 8-bit immediate source.
// Assumes eabyte already scanned.
#define ALU_RM16IMM8(e, expr, stmode, writeback) { \
		register uint32_t dstarg, srcarg, result; \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			srcarg = (int32_t)(int8_t)*inp++; \
			dstarg = e->reg[dstreg]; \
			CHECKINP; \
			result = (expr); \
			if (writeback) PUTREG16(dstreg, result); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			srcarg = (int32_t)(int8_t)*inp++; \
			dstarg = getu16l(e->mem + mea); \
			CHECKINP; \
			result = (expr); \
			if (writeback) putu16l(e->mem + mea, result); \
		} \
		SETST(e, stmode | STM_16, dstarg, srcarg, result); \
	}

#define ALU_RM32IMM8(e, expr, stmode, writeback) { \
		register uint32_t dstarg, srcarg, result; \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			srcarg = (int32_t)(int8_t)*inp++; \
			dstarg = e->reg[dstreg]; \
			CHECKINP; \
			result = (expr); \
			if (writeback) PUTREG32(dstreg, result); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			srcarg = (int32_t)(int8_t)*inp++; \
			dstarg = getu32l(e->mem + mea); \
			CHECKINP; \
			result = (expr); \
			if (writeback) putu32l(e->mem + mea, result); \
		} \
		SETST(e, stmode | STM_32, dstarg, srcarg, result); \
	}


// ALU instruction, AL/AX/EAX dest, immediate source, 2-argument status mode
#define ALU_ALIMM8(e, expr, stmode, writeback) { \
		register uint32_t dstarg = e->reg[EAX]; \
		register uint32_t srcarg = *inp++; \
		CHECKINP; \
		register uint32_t result = (expr); \
		if (writeback) PUTREG8(AL, result); \
		SETST(e, stmode | STM_8, dstarg, srcarg, result); \
	}

#define ALU_AXIMM16(e,expr,stmode, writeback) { \
		register uint32_t dstarg = e->reg[EAX]; \
		register uint32_t srcarg = getu16l(inp); inp += 2; \
		CHECKINP; \
		register uint32_t result = (expr); \
		if (writeback) PUTREG16(AX, result); \
		SETST(e, stmode | STM_16, dstarg, srcarg, result); \
	}

#define ALU_EAXIMM32(e,expr,stmode, writeback) { \
		register uint32_t dstarg = e->reg[EAX]; \
		register uint32_t srcarg = getu32l(inp); inp += 4; \
		CHECKINP; \
		register uint32_t result = (expr); \
		if (writeback) PUTREG32(EAX, result); \
		SETST(e, stmode | STM_32, dstarg, srcarg, result); \
	}


// Unary ALU instruction, r/m dest.
// Assumes eabyte already scanned.
#define ALU_RM8(e, expr, stmode, stdstarg, stsrcarg) { \
		register uint32_t dstarg, result; \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			dstarg = e->reg[dstreg]; \
			CHECKINP; \
			result = (expr); \
			PUTREG8(dstreg, result); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			dstarg = e->mem[mea]; \
			CHECKINP; \
			result = (expr); \
			e->mem[mea] = result; \
		} \
		SETST(e, stmode | STM_8, stdstarg, stsrcarg, result); \
	}
#define ALU_RM16(e, expr, stmode, stdstarg, stsrcarg) { \
		register uint32_t dstarg, result; \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			dstarg = e->reg[dstreg]; \
			CHECKINP; \
			result = (expr); \
			PUTREG16(dstreg, result); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			dstarg = getu16l(e->mem + mea); \
			CHECKINP; \
			result = (expr); \
			putu16l(e->mem + mea, result); \
		} \
		SETST(e, stmode | STM_16, stdstarg, stsrcarg, result); \
	}

#define ALU_RM32(e, expr, stmode, stdstarg, stsrcarg) { \
		register uint32_t dstarg, result; \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			dstarg = e->reg[dstreg]; \
			CHECKINP; \
			result = (expr); \
			PUTREG32(dstreg, result); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			dstarg = getu32l(e->mem + mea); \
			CHECKINP; \
			result = (expr); \
			putu32l(e->mem + mea, result); \
		} \
		SETST(e, stmode | STM_32, stdstarg, stsrcarg, result); \
	}



// Shift/rotate operations
#define DOSHIFT8_16(width,mask,utype,itype) \
	switch (shiftmode) { \
	case 0: { /* ROL */ \
		register unsigned cnt = srcarg & width-1; \
		result = (dstarg << cnt) | \
			((dstarg & mask) >> (width-cnt)); \
		break; } \
	case 1: { /* ROR */ \
		register unsigned cnt = srcarg & width-1; \
		result = ((dstarg & mask) >> cnt) | \
			(dstarg << (width-cnt)); \
		break; } \
	case 2: { /* RCL */ \
		unsigned cf = test_cf(e); \
		shiftmode |= (cf << 3); /* save carry in */ \
		register unsigned cnt = srcarg; \
		while (cnt >= width+1) cnt -= width+1;	\
		result = cnt ? (dstarg << cnt) | (cf << (cnt-1)) | \
					((dstarg & mask) >> (width+1-cnt)) \
			: dstarg; \
		break; } \
	case 3: { /* RCR */ \
		unsigned cf = test_cf(e); \
		shiftmode |= (cf << 3); /* save carry in */ \
		register unsigned cnt = srcarg; \
		while (cnt >= width+1) cnt -= width+1;	\
		result = cnt ? ((dstarg & mask) >> cnt) | \
					(cf << (width-cnt)) | \
					(dstarg << (width+1-cnt)) \
			: dstarg; \
		break; } \
	case 4: { /* SAL/SHL */ \
		result = (dstarg << srcarg); break; } \
	case 5: { /* SHR */ \
		result = ((utype)dstarg >> srcarg); break; } \
	case 6: goto invalid; \
	case 7: { /* SAR */ \
		result = ((itype)dstarg >> srcarg); break; } \
	}

// We have to be a bit more careful with 32-bit shifts
// to avoid using shift counts equal to the promoted C-language operand size.
#define DOSHIFT32 \
	switch (shiftmode) { \
	case 0: { /* ROL */ \
		result = srcarg ? (dstarg << srcarg) | \
					(dstarg >> (32-srcarg)) \
			: dstarg; \
		break; } \
	case 1: { /* ROR */ \
		result = srcarg ? (dstarg >> srcarg) | \
					(dstarg << (32-srcarg)) \
			: dstarg; \
		break; } \
	case 2: { /* RCL */ \
		unsigned cf = test_cf(e); \
		shiftmode |= (cf << 3); /* save carry in */ \
		result = srcarg == 1 ? (dstarg << 1) | cf \
			: srcarg ? (dstarg << srcarg) | (cf << (srcarg-1)) | \
					(dstarg >> (32+1-srcarg)) \
			: dstarg; \
		break; } \
	case 3: { /* RCR */ \
		unsigned cf = test_cf(e); \
		shiftmode |= (cf << 3); /* save carry in */ \
		result = srcarg == 1 ? (dstarg >> 1) | (cf << 31) \
			: srcarg ? (dstarg >> srcarg) | \
					(cf << (32-srcarg)) | \
					(dstarg << (32+1-srcarg)) \
			: dstarg; \
		break; } \
	case 4: { /* SAL/SHL */ \
		result = (dstarg << srcarg); break; } \
	case 5: { /* SHR */ \
		result = ((uint32_t)dstarg >> srcarg); break; } \
	case 6: goto invalid; \
	case 7: { /* SAR */ \
		result = ((int32_t)dstarg >> srcarg); break; } \
	}

#define SHIFT8(srcexpr) { \
		register uint8_t eabyte = *inp++; \
		register uint32_t dstarg, srcarg, result; \
		register unsigned shiftmode = EA_REG(eabyte); \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			srcarg = (srcexpr) & 31; \
			dstarg = e->reg[dstreg]; \
			CHECKINP; \
			DOSHIFT8_16(8, 0xff, uint8_t, int8_t); \
			PUTREG32(dstreg, result); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			srcarg = (srcexpr) & 31; \
			dstarg = e->mem[mea]; \
			CHECKINP; \
			DOSHIFT8_16(8, 0xff, uint8_t, int8_t); \
			e->mem[mea] = result; \
		} \
		SETST(e, STM_SHIFT | shiftmode | STM_8, dstarg, srcarg, result); \
	}

#define SHIFT16(srcexpr) { \
		register uint8_t eabyte = *inp++; \
		register uint32_t dstarg, srcarg, result; \
		register unsigned shiftmode = EA_REG(eabyte); \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			srcarg = (srcexpr) & 31; \
			dstarg = e->reg[dstreg]; \
			CHECKINP; \
			DOSHIFT8_16(16, 0xffff, uint16_t, int16_t); \
			e->reg[dstreg] = (dstarg & 0xffff0000) | \
					(result & 0x0000ffff); \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			srcarg = (srcexpr) & 31; \
			dstarg = e->mem[mea]; \
			CHECKINP; \
			DOSHIFT8_16(16, 0xffff, uint16_t, int16_t); \
			e->mem[mea] = result; \
		} \
		SETST(e, STM_SHIFT | shiftmode | STM_8, dstarg, srcarg, result); \
	}

#define SHIFT32(srcexpr) { \
		register uint8_t eabyte = *inp++; \
		register uint32_t dstarg, srcarg, result; \
		register unsigned shiftmode = EA_REG(eabyte); \
		if (EA_ISREG(eabyte)) { \
			register unsigned dstreg = EA_RM(eabyte); \
			srcarg = (srcexpr) & 31; \
			dstarg = e->reg[dstreg]; \
			CHECKINP; \
			DOSHIFT32; \
			e->reg[dstreg] = result; \
		} else { \
			CHECKMEA(e, eabyte, inp, mea); \
			srcarg = (srcexpr) & 31; \
			dstarg = e->mem[mea]; \
			CHECKINP; \
			DOSHIFT32; \
			e->mem[mea] = result; \
		} \
		SETST(e, STM_SHIFT | shiftmode | STM_8, dstarg, srcarg, result); \
	}


// Special immediate-to-register MOV instruction forms
#define MOV_REGIMM8(r) { \
		uint32_t srcarg = *inp++; \
		CHECKINP; \
		PUTREG8(r, srcarg); \
	}

#define MOV_REGIMM16(r) { \
		uint32_t srcarg = getu16l(inp); inp += 2; \
		CHECKINP; \
		PUTREG16(r, srcarg); \
	}

#define MOV_REGIMM32(r) { \
		uint32_t srcarg = getu32l(inp); inp += 4; \
		CHECKINP; \
		PUTREG32(r, srcarg); \
	}


#define PUSH(val) { \
		CHECKINP; \
		register uint32_t mea = e->reg[ESP] - 4; \
		if (mea >= e->memsize) assert(0); /* XXX */ \
		putu32l(e->mem + mea, val); \
		e->reg[ESP] = mea; \
	}

#define POP(val) { \
		CHECKINP; \
		register uint32_t mea = e->reg[ESP]; \
		if (mea >= e->memsize) assert(0); /* XXX */ \
		val = getu32l(e->mem + mea); \
		e->reg[ESP] = mea + 4; \
	}


// JMP/Jcc/CALL instructions
#define JUMP(target) { \
		e->eip = (target); \
		goto neweip; \
	}

#define JUMP_DISP8(cond) { \
		register int32_t target = (int8_t)*inp++; \
		CHECKINP; \
		if (cond) { \
			target += (inp - e->mem); \
			JUMP(target); \
		} \
	}

#define JUMP_DISP32(cond, iscall) { \
		register uint32_t target = getu32l(inp); inp += 4; \
		CHECKINP; \
		if (cond) { \
			target += (inp - e->mem); \
if (iscall) printf("CALL from %08x esp=%08x ebp=%08x\n", inp - e->mem, e->reg[ESP], e->reg[EBP]); \
			if (iscall) PUSH(inp - e->mem); \
			JUMP(target); \
		} \
	}


#define INC_REG16(r) { \
		CHECKINP; \
		PUTREG16(r, e->reg[r] + 1); \
	}
#define INC_REG32(r) { \
		CHECKINP; \
		PUTREG32(r, e->reg[r] + 1); \
	}

#define DEC_REG16(r) { \
		CHECKINP; \
		PUTREG16(r, e->reg[r] - 1); \
	}
#define DEC_REG32(r) { \
		CHECKINP; \
		PUTREG32(r, e->reg[r] - 1); \
	}

#define XCHG_REGREG(r1,r2) { \
		CHECKINP; \
		uint32_t tmp = e->reg[r1]; \
		e->reg[r1] = e->reg[r2]; \
		e->reg[r2] = tmp; \
	}

#define BSWAP_REG32(r)	CHECKINP; e->reg[r] = bswap32(e->reg[r]);

#define SET_RM8(cond) { \
		register uint8_t eabyte = *inp++; \
		ALU_RM8(e, (cond) ? 0xff : 0, 0, 0, 0); \
	}


// Translation pass 1:
// scan instruction stream, build preliminary vxinsn table,
// and decide how many instructions to translate in this fragment.
static int emulate(vxemu *e)
{
neweip: ;
	// Find and check permissions on the input instruction stream,
	// and determine how far ahead we can scan (up to one full page)
	// before hitting a non-executable page.
	uint8_t *instart = e->mem + e->eip;
	if (!(checkperm(e, e->eip) & VXPERM_EXEC)) {
		e->trapno = VXTRAP_PAGEFAULT;
		e->traperr = 0x10;		// Execute fault
		e->trapva = e->eip;
		return VXTRAP_PAGEFAULT;
	}
	uint8_t *inmax = (uint8_t*)e->mem +
				VXPAGETRUNC(e->eip) + VXPAGESIZE;
	if (checkperm(e, e->eip + VXPAGESIZE) & VXPERM_EXEC)
		inmax += VXPAGESIZE;

	uint8_t *inp = instart;
	while (1) {
		// Begin instruction decode.
		// We might take a fault on any of these instruction reads
		// if we run off the end of a mapped code page.
		// In that case our exception handler
		// notices that e->ininst != NULL and initiates recovery.
		// Or we might _not_ take a fault
		// on a page marked read-only but not executable;
		// that's why we check each insn's final inp against inmax
		// before committing the instruction's results.
printf("%08x esp=%08x\n", inp - e->mem, e->reg[ESP]);
		e->ininst = inp;
		switch (*inp++) {

		// ADD
		case 0x00:				// ADD rm8,r8
			ALU_RMREG8(e, dstarg + srcarg, STM_ADD, 1);
			break;
		case 0x01:				// ADD rm32,r32
			ALU_RMREG32(e, dstarg + srcarg, STM_ADD, 1);
			break;
		case 0x02:				// ADD r8,rm8
			ALU_REGRM8(e, dstarg + srcarg, STM_ADD, 1);
			break;
		case 0x03:				// ADD r32,rm32
			ALU_REGRM32(e, dstarg + srcarg, STM_ADD, 1);
			break;
		case 0x04:				// ADD AL,imm8
			ALU_ALIMM8(e, dstarg + srcarg, STM_ADD, 1);
			break;
		case 0x05:				// ADD EAX,imm32
			ALU_EAXIMM32(e, dstarg + srcarg, STM_ADD, 1);
			break;

		// ADC
		case 0x10: {				// ADC rm8,r8
			unsigned cf = test_cf(e);
			ALU_RMREG8(e, dstarg + srcarg + cf, STM_ADD | cf, 1);
			break; }
		case 0x11: {				// ADC rm32,r32
			unsigned cf = test_cf(e);
			ALU_RMREG32(e, dstarg + srcarg + cf, STM_ADD | cf, 1);
			break; }
		case 0x12: {				// ADC r8,rm8
			unsigned cf = test_cf(e);
			ALU_REGRM8(e, dstarg + srcarg + cf, STM_ADD | cf, 1);
			break; }
		case 0x13: {				// ADC r32,rm32
			unsigned cf = test_cf(e);
			ALU_REGRM32(e, dstarg + srcarg + cf, STM_ADD | cf, 1);
			break; }
		case 0x14: {				// ADC AL,imm8
			unsigned cf = test_cf(e);
			ALU_ALIMM8(e, dstarg + srcarg + cf, STM_ADD | cf, 1);
			break; }
		case 0x15: {				// ADC EAX,imm32
			unsigned cf = test_cf(e);
			ALU_EAXIMM32(e, dstarg + srcarg + cf, STM_ADD | cf, 1);
			break; }

		// SUB
		case 0x28:				// SUB rm8,r8
			ALU_RMREG8(e, dstarg - srcarg, STM_SUB, 1);
			break;
		case 0x29:				// SUB rm32,r32
			ALU_RMREG32(e, dstarg - srcarg, STM_SUB, 1);
			break;
		case 0x2a:				// SUB r8,rm8
			ALU_REGRM8(e, dstarg - srcarg, STM_SUB, 1);
			break;
		case 0x2b:				// SUB r32,rm32
			ALU_REGRM32(e, dstarg - srcarg, STM_SUB, 1);
			break;
		case 0x2c:				// SUB AL,imm8
			ALU_ALIMM8(e, dstarg - srcarg, STM_SUB, 1);
			break;
		case 0x2d:				// SUB EAX,imm32
			ALU_EAXIMM32(e, dstarg - srcarg, STM_SUB, 1);
			break;

		// SBB
		case 0x18: {				// SBB rm8,r8
			unsigned cf = test_cf(e);
			ALU_RMREG8(e, dstarg - srcarg - cf, STM_SUB | cf, 1);
			break; }
		case 0x19: {				// SBB rm32,r32
			unsigned cf = test_cf(e);
			ALU_RMREG32(e, dstarg - srcarg - cf, STM_SUB | cf, 1);
			break; }
		case 0x1a: {				// SBB r8,rm8
			unsigned cf = test_cf(e);
			ALU_REGRM8(e, dstarg - srcarg - cf, STM_SUB | cf, 1);
			break; }
		case 0x1b: {				// SBB r32,rm32
			unsigned cf = test_cf(e);
			ALU_REGRM32(e, dstarg - srcarg - cf, STM_SUB | cf, 1);
			break; }
		case 0x1c: {				// SBB AL,imm8
			unsigned cf = test_cf(e);
			ALU_ALIMM8(e, dstarg - srcarg - cf, STM_SUB | cf, 1);
			break; }
		case 0x1d: {				// SBB EAX,imm32
			unsigned cf = test_cf(e);
			ALU_EAXIMM32(e, dstarg - srcarg - cf, STM_SUB | cf, 1);
			break; }

		// CMP
		case 0x38:				// CMP rm8,r8
			ALU_RMREG8(e, dstarg - srcarg, STM_SUB, 0);
			break;
		case 0x39:				// CMP rm32,r32
			ALU_RMREG32(e, dstarg - srcarg, STM_SUB, 0);
			break;
		case 0x3a:				// CMP r8,rm8
			ALU_REGRM8(e, dstarg - srcarg, STM_SUB, 0);
			break;
		case 0x3b:				// CMP r32,rm32
			ALU_REGRM32(e, dstarg - srcarg, STM_SUB, 0);
			break;
		case 0x3c:				// CMP AL,imm8
			ALU_ALIMM8(e, dstarg - srcarg, STM_SUB, 0);
			break;
		case 0x3d:				// CMP EAX,imm32
			ALU_EAXIMM32(e, dstarg - srcarg, STM_SUB, 0);
			break;

		// AND
		case 0x20:				// AND rm8,r8
			ALU_RMREG8(e, dstarg & srcarg, STM_LOG, 1);
			break;
		case 0x21:				// AND rm32,r32
			ALU_RMREG32(e, dstarg & srcarg, STM_LOG, 1);
			break;
		case 0x22:				// AND r8,rm8
			ALU_REGRM8(e, dstarg & srcarg, STM_LOG, 1);
			break;
		case 0x23:				// AND r32,rm32
			ALU_REGRM32(e, dstarg & srcarg, STM_LOG, 1);
			break;
		case 0x24:				// AND AL,imm8
			ALU_ALIMM8(e, dstarg & srcarg, STM_LOG, 1);
			break;
		case 0x25:				// AND EAX,imm32
			ALU_EAXIMM32(e, dstarg & srcarg, STM_LOG, 1);
			break;

		// OR
		case 0x08:				// OR rm8,r8
			ALU_RMREG8(e, dstarg | srcarg, STM_LOG, 1);
			break;
		case 0x09:				// OR rm32,r32
			ALU_RMREG32(e, dstarg | srcarg, STM_LOG, 1);
			break;
		case 0x0a:				// OR r8,rm8
			ALU_REGRM8(e, dstarg | srcarg, STM_LOG, 1);
			break;
		case 0x0b:				// OR r32,rm32
			ALU_REGRM32(e, dstarg | srcarg, STM_LOG, 1);
			break;
		case 0x0c:				// OR AL,imm8
			ALU_ALIMM8(e, dstarg | srcarg, STM_LOG, 1);
			break;
		case 0x0d:				// OR EAX,imm32
			ALU_EAXIMM32(e, dstarg | srcarg, STM_LOG, 1);
			break;

		// XOR
		case 0x30:				// XOR rm8,r8
			ALU_RMREG8(e, dstarg ^ srcarg, STM_LOG, 1);
			break;
		case 0x31:				// XOR rm32,r32
			ALU_RMREG32(e, dstarg ^ srcarg, STM_LOG, 1);
			break;
		case 0x32:				// XOR r8,rm8
			ALU_REGRM8(e, dstarg ^ srcarg, STM_LOG, 1);
			break;
		case 0x33:				// XOR r32,rm32
			ALU_REGRM32(e, dstarg ^ srcarg, STM_LOG, 1);
			break;
		case 0x34:				// XOR AL,imm8
			ALU_ALIMM8(e, dstarg ^ srcarg, STM_LOG, 1);
			break;
		case 0x35:				// XOR EAX,imm32
			ALU_EAXIMM32(e, dstarg ^ srcarg, STM_LOG, 1);
			break;

		// TEST
		case 0x84:				// TEST rm8,r8
			ALU_RMREG8(e, dstarg & srcarg, STM_LOG, 0);
			break;
		case 0x85:				// TEST rm32,r32
			ALU_RMREG32(e, dstarg & srcarg, STM_LOG, 0);
			break;
		case 0xa8:				// XOR AL,imm8
			ALU_ALIMM8(e, dstarg & srcarg, STM_LOG, 0);
			break;
		case 0xa9:				// XOR EAX,imm32
			ALU_EAXIMM32(e, dstarg & srcarg, STM_LOG, 0);
			break;

		// MOV
		case 0x88:				// MOV rm8,r8
			ALU_RMREG8(e, srcarg, 0, 1); break;
		case 0x89:				// MOV rm32,r32
			ALU_RMREG32(e, srcarg, 0, 1); break;
		case 0x8a:				// MOV r8,rm8
			ALU_REGRM8(e, srcarg, 0, 1); break;
		case 0x8b:				// MOV r32,rm32
			ALU_REGRM32(e, srcarg, 0, 1); break;
		case 0xa0: {				// MOV AL,moffs8
			uint32_t mea = getu32l(inp); inp += 4;
			CHECKINP;
			if (mea >= e->memsize) assert(0); /* XXX */
			e->reg[EAX] = (e->reg[EAX] & 0xffffff00) |
					e->mem[mea];
			break;
		    }
		case 0xa1: {				// MOV EAX,moffs32
			uint32_t mea = getu32l(inp); inp += 4;
			CHECKINP;
			if (mea >= e->memsize) assert(0); /* XXX */
			e->reg[EAX] = getu32l(e->mem + mea);
			break;
		    }
		case 0xa2: {				// MOV moffs8,AL
			uint32_t mea = getu32l(inp); inp += 4;
			CHECKINP;
			if (mea >= e->memsize) assert(0); /* XXX */
			e->mem[mea] = e->reg[EAX];
			break;
		    }
		case 0xa3: {				// MOV moffs32,EAX
			uint32_t mea = getu32l(inp); inp += 4;
			CHECKINP;
			if (mea >= e->memsize) assert(0); /* XXX */
			putu32l(e->mem + mea, e->reg[EAX]);
			break;
		    }
		case 0xb0:				// MOV AL,imm8
			MOV_REGIMM8(AL); break;
		case 0xb1:				// MOV CL,imm8
			MOV_REGIMM8(CL); break;
		case 0xb2:				// MOV DL,imm8
			MOV_REGIMM8(DL); break;
		case 0xb3:				// MOV BL,imm8
			MOV_REGIMM8(BL); break;
		case 0xb4:				// MOV AH,imm8
			MOV_REGIMM8(AH); break;
		case 0xb5:				// MOV CH,imm8
			MOV_REGIMM8(DH); break;
		case 0xb6:				// MOV DH,imm8
			MOV_REGIMM8(CH); break;
		case 0xb7:				// MOV BH,imm8
			MOV_REGIMM8(BH); break;
		case 0xb8:				// MOV EAX,imm32
			MOV_REGIMM32(EAX); break;
		case 0xb9:				// MOV ECX,imm32
			MOV_REGIMM32(ECX); break;
		case 0xba:				// MOV EDX,imm32
			MOV_REGIMM32(EDX); break;
		case 0xbb:				// MOV EBX,imm32
			MOV_REGIMM32(EBX); break;
		case 0xbc:				// MOV ESP,imm32
			MOV_REGIMM32(ESP); break;
		case 0xbd:				// MOV EBP,imm32
			MOV_REGIMM32(EBP); break;
		case 0xbe:				// MOV ESI,imm32
			MOV_REGIMM32(ESI); break;
		case 0xbf:				// MOV EDI,imm32
			MOV_REGIMM32(EDI); break;
		case 0xc6: {				// MOV rm8,imm8
			register uint8_t eabyte = *inp++;
			ALU_RMIMM8(e, srcarg, 0, 1);
			break;
		    }
		case 0xc7: {				// MOV rm32,imm32
			register uint8_t eabyte = *inp++;
			ALU_RMIMM32(e, srcarg, 0, 1); break;
			break;
		    }


		// XCHG
		case 0x90:				// XCHG EAX,EAX (NOP)
			break;
		case 0x91:				// XCHG ECX,EAX
			XCHG_REGREG(EAX,ECX); break;
		case 0x92:				// XCHG EDX,EAX
			XCHG_REGREG(EAX,EDX); break;
		case 0x93:				// XCHG EBX,EAX
			XCHG_REGREG(EAX,EBX); break;
		case 0x94:				// XCHG ESP,EAX
			XCHG_REGREG(EAX,ESP); break;
		case 0x95:				// XCHG EBP,EAX
			XCHG_REGREG(EAX,EBP); break;
		case 0x96:				// XCHG ESI,EAX
			XCHG_REGREG(EAX,ESI); break;
		case 0x97:				// XCHG EDI,EAX
			XCHG_REGREG(EAX,EDI); break;
		case 0x86: {				// XCHG rm8,r8
			register uint8_t eabyte = *inp++;
			register uint32_t rreg = EA_REG(eabyte);
			register uint32_t rval = e->reg[rreg];
			if (EA_ISREG(eabyte)) {
				register unsigned rmreg = EA_RM(eabyte);
				CHECKINP;
				PUTREG8(rreg, e->reg[rmreg]);
				PUTREG8(rmreg, rval);
			} else {
				CHECKMEA(e, eabyte, inp, mea);
				CHECKINP;
				register uint32_t rmval = e->mem[mea];
				e->mem[mea] = rval;
				PUTREG8(rreg, rmval);	// AFTER memory write!
			}
			break;
		    }
		case 0x87: {				// XCHG rm32,r32
			register uint8_t eabyte = *inp++;
			register uint32_t rreg = EA_REG(eabyte);
			register uint32_t rval = e->reg[rreg];
			if (EA_ISREG(eabyte)) {
				register unsigned rmreg = EA_RM(eabyte);
				CHECKINP;
				e->reg[rreg] = e->reg[rmreg];
				e->reg[rmreg] = rval;
			} else {
				CHECKMEA(e, eabyte, inp, mea);
				CHECKINP;
				register uint32_t rmval = getu32l(e->mem + mea);
				putu32l(e->mem + mea, rval);
				e->reg[rreg] = rmval;	// AFTER memory write!
			}
			break;
		    }

		// LEA r32,m
		case 0x8d: {
			register uint8_t eabyte = *inp++;
			if (EA_ISREG(eabyte))
				goto invalid;	// Mem-only!
			uint32_t mea = scanmea(e, eabyte, &inp);
			CHECKINP;
			e->reg[EA_REG(eabyte)] = mea;
			break;
		    }


		// OP Eb,Ib; OP Ev,Ib; IMUL Gv,Ev,Ib
		case 0x80: {				// OP rm8,imm8
			register uint8_t eabyte = *inp++;
			switch (EA_REG(eabyte)) {
			case 0:				// ADD rm8,imm8
				ALU_RMIMM8(e, dstarg + srcarg, STM_ADD, 1);
				break;
			case 1:				// OR rm8,imm8
				ALU_RMIMM8(e, dstarg | srcarg, STM_LOG, 1);
				break;
			case 2: {			// ADC rm8,imm8
				unsigned cf = test_cf(e);
				ALU_RMIMM8(e, dstarg + srcarg + cf,
						STM_ADD | cf, 1);
				break; }
			case 3:	{			// SBB rm8,imm8
				unsigned cf = test_cf(e);
				ALU_RMIMM8(e, dstarg - srcarg - cf,
						STM_SUB | cf, 1);
				break; }
			case 4:				// AND rm8,imm8
				ALU_RMIMM8(e, dstarg & srcarg, STM_LOG, 1);
				break;
			case 5:				// SUB rm8,imm8
				ALU_RMIMM8(e, dstarg - srcarg, STM_SUB, 1);
				break;
			case 6:				// XOR rm8,imm8
				ALU_RMIMM8(e, dstarg ^ srcarg, STM_LOG, 1);
				break;
			case 7:				// CMP rm8,imm8
				ALU_RMIMM8(e, dstarg - srcarg, STM_SUB, 0);
				break;
			}
			break;
		    }

		case 0x81: {				// OP rm32,imm32
			register uint8_t eabyte = *inp++;
			switch (EA_REG(eabyte)) {
			case 0:				// ADD rm8,imm8
				ALU_RMIMM32(e, dstarg + srcarg, STM_ADD, 1);
				break;
			case 1:				// OR rm8,imm8
				ALU_RMIMM32(e, dstarg | srcarg, STM_LOG, 1);
				break;
			case 2:	{			// ADC rm8,imm8
				unsigned cf = test_cf(e);
				ALU_RMIMM32(e, dstarg + srcarg + cf,
						STM_ADD | cf, 1);
				break; }
			case 3:	{			// SBB rm8,imm8
				unsigned cf = test_cf(e);
				ALU_RMIMM32(e, dstarg - srcarg - cf,
						STM_SUB | cf, 1);
				break; }
			case 4:				// AND rm8,imm8
				ALU_RMIMM32(e, dstarg & srcarg, STM_LOG, 1);
				break;
			case 5:				// SUB rm8,imm8
				ALU_RMIMM32(e, dstarg - srcarg, STM_SUB, 1);
				break;
			case 6:				// XOR rm8,imm8
				ALU_RMIMM32(e, dstarg ^ srcarg, STM_LOG, 1);
				break;
			case 7:				// CMP rm8,imm8
				ALU_RMIMM32(e, dstarg - srcarg, STM_SUB, 0);
				break;
			}
			break;
		    }

		case 0x83: {				// OP rm32,imm8
			register uint8_t eabyte = *inp++;
			switch (EA_REG(eabyte)) {
			case 0:				// ADD rm8,imm8
				ALU_RM32IMM8(e, dstarg + srcarg, STM_ADD, 1);
				break;
			case 1:				// OR rm8,imm8
				ALU_RM32IMM8(e, dstarg | srcarg, STM_LOG, 1);
				break;
			case 2:	{			// ADC rm8,imm8
				unsigned cf = test_cf(e);
				ALU_RM32IMM8(e, dstarg + srcarg + cf,
						STM_ADD | cf, 1);
				break; }
			case 3:	{			// SBB rm8,imm8
				unsigned cf = test_cf(e);
				ALU_RM32IMM8(e, dstarg - srcarg - cf,
						STM_SUB | cf, 1);
				break; }
			case 4:				// AND rm8,imm8
				ALU_RM32IMM8(e, dstarg & srcarg, STM_LOG, 1);
				break;
			case 5:				// SUB rm8,imm8
				ALU_RM32IMM8(e, dstarg - srcarg, STM_SUB, 1);
				break;
			case 6:				// XOR rm8,imm8
				ALU_RM32IMM8(e, dstarg ^ srcarg, STM_LOG, 1);
				break;
			case 7:				// CMP rm8,imm8
				ALU_RM32IMM8(e, dstarg - srcarg, STM_SUB, 0);
				break;
			}
			break;
		    }

		case 0xf6: {				// OP rm8,imm8
			register uint8_t eabyte = *inp++;
			switch (EA_REG(eabyte)) {
			case 0:				// TEST rm8,imm8
				ALU_RMIMM8(e, dstarg & srcarg, STM_LOG, 0);
				break;
			case 2:				// NOT rm8
				ALU_RM8(e, ~dstarg, 0, 0, 0);
				break;
			case 3:				// NEG rm8
				// Set flags to reflect a '0-dstarg' subtract
				ALU_RM8(e, -dstarg, STM_SUB, 0, dstarg);
				break;
			case 4:	{			// MUL rm8
				LOADRM8(e, eabyte, inp, srcarg);
				CHECKINP;
				register uint32_t dstarg = e->reg[EAX];
				uint32_t result = (uint8_t)dstarg * srcarg;
				PUTREG16(AX, result);
				SETST(e, STM_MUL | STM_8,
					dstarg, srcarg, result);
				break;
			    }
			case 5:	{			// IMUL rm8
				LOADRM8(e, eabyte, inp, srcarg);
				CHECKINP;
				register uint32_t dstarg = e->reg[EAX];
				register int32_t result = 
					(int8_t)dstarg * (int8_t)srcarg;
				PUTREG16(AX, result);
				SETST(e, STM_IMUL | STM_8,
					dstarg, srcarg, result);
				break;
			    }
			case 6: {			// DIV rm8
				LOADRM8(e, eabyte, inp, srcarg);
				CHECKINP;
				uint32_t dstarg = e->reg[EAX];
				register unsigned quotient =
					(uint16_t)dstarg / (uint8_t)srcarg;
				register unsigned remainder =
					(uint16_t)dstarg % (uint8_t)srcarg;
				if ((quotient >> 8) != 0)
					goto diverr;
				PUTREG16(AX, (remainder << 8) | quotient);
				SETST(e, STM_LOG | STM_8,
					dstarg, srcarg, quotient);
				break;
			    }
			case 7: {			// IDIV rm8
				LOADRM8(e, eabyte, inp, srcarg);
				CHECKINP;
				uint32_t dstarg = e->reg[EAX];
				register int16_t quotient =
					(int16_t)dstarg / (int8_t)srcarg;
				register int16_t remainder =
					(int16_t)dstarg % (int8_t)srcarg;
				if (quotient < -128 || quotient > 127)
					goto diverr;
				PUTREG16(AX, ((remainder & 0xff) << 8) |
						(quotient & 0xff));
				SETST(e, STM_LOG | STM_8,
					dstarg, srcarg, quotient);
				break;
			    }
			default:
				goto invalid;
			}
			break;
		    }

		case 0xf7: {				// OP rm32,imm32
			register uint8_t eabyte = *inp++;
			switch (EA_REG(eabyte)) {
			case 0:				// TEST rm32,imm32
				ALU_RMIMM32(e, dstarg & srcarg, STM_LOG, 0);
				break;
			case 2:				// NOT rm32
				ALU_RM32(e, ~dstarg, 0, 0, 0);
				break;
			case 3:				// NEG rm32
				// Set flags to reflect a '0-dstarg' subtract
				ALU_RM32(e, -dstarg, STM_SUB, 0, dstarg);
				break;
			case 4:	{			// MUL rm32
				LOADRM32(e, eabyte, inp, srcarg);
				CHECKINP;
				register uint32_t dstarg = e->reg[EAX];
				uint64_t result = dstarg * srcarg;
				e->reg[EAX] = result;
				e->reg[EDX] = result >> 32;
				SETST(e, STM_MUL | STM_32,
					dstarg, srcarg, result);
				break; }
			case 5:	{			// IMUL rm32
				LOADRM32(e, eabyte, inp, srcarg);
				CHECKINP;
				register uint32_t dstarg = e->reg[EAX];
				int64_t result =
					(int32_t)dstarg * (int32_t)srcarg;
				e->reg[EAX] = result;
				e->reg[EDX] = result >> 32;
				SETST(e, STM_MUL | STM_32,
					dstarg, srcarg, result);
				break; }
			case 6: {			// DIV rm32
				LOADRM32(e, eabyte, inp, srcarg);
				CHECKINP;
				register uint64_t dividend =
					((uint64_t)e->reg[EAX]) |
					((uint64_t)e->reg[EDX] << 32);
				register uint64_t quotient = dividend / srcarg;
				register uint32_t remainder = dividend % srcarg;
				if ((quotient >> 32) != 0)
					goto diverr;
				e->reg[EAX] = quotient;
				e->reg[EDX] = remainder;
				SETST(e, STM_MUL | STM_32,
					dividend, srcarg, quotient);
				break; }
			case 7: {			// IDIV rm32
				LOADRM32(e, eabyte, inp, srcarg);
				CHECKINP;
				register int64_t dividend =
					((int64_t)e->reg[EAX]) |
					((int64_t)e->reg[EDX] << 32);
				register int64_t quotient =
					dividend / (int32_t)srcarg;
				register int32_t remainder =
					dividend % (int32_t)srcarg;
				if (((int32_t)quotient >> 31) != 
						(int32_t)(quotient >> 32))
					goto diverr;
				e->reg[EAX] = quotient;
				e->reg[EDX] = remainder;
				SETST(e, STM_MUL | STM_32,
					dividend, srcarg, quotient);
				break; }
			default:
				goto invalid;
			}
			break;
		    }

		// INC/DEC
		case 0x40: INC_REG32(EAX); break;	// INC EAX
		case 0x41: INC_REG32(ECX); break;	// INC ECX
		case 0x42: INC_REG32(EDX); break;	// INC EDX
		case 0x43: INC_REG32(EBX); break;	// INC EBX
		case 0x44: INC_REG32(ESP); break;	// INC ESP
		case 0x45: INC_REG32(EBP); break;	// INC EBP
		case 0x46: INC_REG32(ESI); break;	// INC ESI
		case 0x47: INC_REG32(EDI); break;	// INC EDI

		case 0x48: DEC_REG32(EAX); break;	// DEC EAX
		case 0x49: DEC_REG32(ECX); break;	// DEC ECX
		case 0x4a: DEC_REG32(EDX); break;	// DEC EDX
		case 0x4b: DEC_REG32(EBX); break;	// DEC EBX
		case 0x4c: DEC_REG32(ESP); break;	// DEC ESP
		case 0x4d: DEC_REG32(EBP); break;	// DEC EBP
		case 0x4e: DEC_REG32(ESI); break;	// DEC ESI
		case 0x4f: DEC_REG32(EDI); break;	// DEC EDI


		// IMUL (3-operand forms)
		case 0x6b: {				// IMUL r32,rm32,imm8
			register uint8_t eabyte = *inp++;
			LOADRM32(e, eabyte, inp, src1arg);
			register int32_t src2arg = (int8_t)*inp++;
			CHECKINP;
			int32_t result = (int32_t)src1arg * (int32_t)src2arg;
			e->reg[EA_REG(eabyte)] = result;
			SETST(e, STM_IMUL | STM_32, src1arg, src2arg, result);
			break;
		    }
		case 0x69: {				// IMUL r32,rm32,imm32
			register uint8_t eabyte = *inp++;
			LOADRM32(e, eabyte, inp, src1arg);
			register int32_t src2arg = getu32l(inp); inp += 4;
			CHECKINP;
			int32_t result = (int32_t)src1arg * (int32_t)src2arg;
			e->reg[EA_REG(eabyte)] = result;
			SETST(e, STM_IMUL | STM_32, src1arg, src2arg, result);
			break;
		    }


		// Shift/Rotate
		case 0xd0:				// OP rm8,1
			SHIFT8(1); break;
		case 0xd1:				// OP rm32,1
			SHIFT32(1); break;
		case 0xd2:				// OP rm8,CL
			SHIFT8(e->reg[ECX]); break;
		case 0xd3:				// OP rm32,CL
			SHIFT32(e->reg[ECX]); break;
		case 0xc0:				// OP rm8,imm8
			SHIFT8(*inp++); break;
		case 0xc1:				// OP rm32,imm8
			SHIFT32(*inp++); break;


		// Stack and control flow instructions
		case 0x50: PUSH(e->reg[EAX]); break;	// PUSH r32
		case 0x51: PUSH(e->reg[ECX]); break;
		case 0x52: PUSH(e->reg[EDX]); break;
		case 0x53: PUSH(e->reg[EBX]); break;
		case 0x54: PUSH(e->reg[ESP]); break;
		case 0x55: PUSH(e->reg[EBP]); break;
		case 0x56: PUSH(e->reg[ESI]); break;
		case 0x57: PUSH(e->reg[EDI]); break;
		case 0x6a: {				// PUSH imm8
			int32_t srcarg = (int8_t)*inp++;
			PUSH((int8_t)*inp++);
			break;
		    }
		case 0x68: {				// PUSH imm32
			int32_t srcarg = getu32l(inp); inp += 4;
			PUSH(getu32l(inp));
			break;
		    }

		case 0x58: POP(e->reg[EAX]); break;	// POP r32
		case 0x59: POP(e->reg[ECX]); break;
		case 0x5a: POP(e->reg[EDX]); break;
		case 0x5b: POP(e->reg[EBX]); break;
		case 0x5c: POP(e->reg[ESP]); break;
		case 0x5d: POP(e->reg[EBP]); break;
		case 0x5e: POP(e->reg[ESI]); break;
		case 0x5f: POP(e->reg[EDI]); break;

		case 0xeb: JUMP_DISP8(1);		// JMP rel8
		case 0xe9: JUMP_DISP32(1, 0);		// JMP rel32
		case 0xe8: JUMP_DISP32(1, 1);		// CALL rel32

		case 0x70: JUMP_DISP8(test_of(e)); break;	// JO rel8
		case 0x71: JUMP_DISP8(!test_of(e)); break;	// JNO rel8
		case 0x72: JUMP_DISP8(test_cf(e)); break;	// JC rel8
		case 0x73: JUMP_DISP8(!test_cf(e)); break;	// JNC rel8
		case 0x74: JUMP_DISP8(test_zf(e)); break;	// JZ rel8
		case 0x75: JUMP_DISP8(!test_zf(e)); break;	// JNZ rel8
		case 0x76: JUMP_DISP8(test_be(e)); break;	// JBE rel8
		case 0x77: JUMP_DISP8(!test_be(e)); break;	// JNBE rel8
		case 0x78: JUMP_DISP8(test_sf(e)); break;	// JS rel8
		case 0x79: JUMP_DISP8(!test_sf(e)); break;	// JNS rel8
		case 0x7a: JUMP_DISP8(test_pf(e)); break;	// JP rel8
		case 0x7b: JUMP_DISP8(!test_pf(e)); break;	// JNP rel8
		case 0x7c: JUMP_DISP8(test_l(e)); break;	// JL rel8
		case 0x7d: JUMP_DISP8(!test_l(e)); break;	// JNL rel8
		case 0x7e: JUMP_DISP8(test_le(e)); break;	// JLE rel8
		case 0x7f: JUMP_DISP8(!test_le(e)); break;	// JNLE rel8

		case 0x8f: {
			register uint8_t eabyte = *inp++;
			switch (EA_REG(eabyte)) {
			case 0:				// POP rm32
				if (EA_ISREG(eabyte)) {
					POP(e->reg[EA_RM(eabyte)]);
				} else {
					CHECKMEA(e, eabyte, inp, mea);
					CHECKINP;
					register uint32_t esp = e->reg[ESP];
					if (esp >= e->memsize) assert(0); //XXX
					putu32(e->mem + mea,
						getu32(e->mem + esp));
					e->reg[ESP] = esp + 4; // AFTER write!
				}
				break;

			default:
				goto invalid;
			}
			break;
		    }

		case 0xfe: {
			register uint8_t eabyte = *inp++;
			switch (EA_REG(eabyte)) {
			case 0:				// INC rm8
				// set flags to reflect a 'dstarg+1' add
				ALU_RM8(e, dstarg + 1, STM_ADD, dstarg, 1);
				break;

			case 1:				// DEC rm8
				// set flags to reflect a 'dstarg-1' subtract
				ALU_RM8(e, dstarg - 1, STM_SUB, dstarg, 1);
				break;

			default:
				goto invalid;
			}
			break;
		}

		case 0xff: {
			register uint8_t eabyte = *inp++;
			switch (EA_REG(eabyte)) {
			case 0:				// INC rm32
				// set flags to reflect a 'dstarg+1' add
				ALU_RM32(e, dstarg + 1, STM_ADD, dstarg, 1);
				break;

			case 1:				// DEC rm32
				// set flags to reflect a 'dstarg-1' subtract
				ALU_RM32(e, dstarg - 1, STM_SUB, dstarg, 1);
				break;

			case 2:	{			// CALL rm32
				LOADRM32(e, eabyte, inp, target);
printf("CALL from %08x esp=%08x ebp=%08x\n", inp - e->mem, e->reg[ESP], e->reg[EBP]);
				PUSH(inp - e->mem);
				JUMP(target);
				break;
			    }

			case 4: {			// JMP rm32
				LOADRM32(e, eabyte, inp, target);
				CHECKINP;
				JUMP(target);
				break;
			    }

			case 6:				// PUSH rm32
				if (EA_ISREG(eabyte)) {
					PUSH(e->reg[EA_RM(eabyte)]);
				} else {
					CHECKMEA(e, eabyte, inp, mea); \
					CHECKINP;
					register uint32_t esp = e->reg[ESP] - 4;
					if (esp >= e->memsize) assert(0); //XXX
					putu32(e->mem + esp,
						getu32(e->mem + mea));
					e->reg[ESP] = esp;	// AFTER write!
				}
				break;

			default:
				goto invalid;
			}
			break;
		    }

		// CS and DS segment overrides, only valid for branch hints
		case 0x2e:	// CS/"not taken"
		case 0x3e:	// DS/"taken"
			switch (*inp++) {

			// Jcc (8-bit displacement)
			case 0x70: JUMP_DISP8(test_of(e)); break;
			case 0x71: JUMP_DISP8(!test_of(e)); break;
			case 0x72: JUMP_DISP8(test_cf(e)); break;
			case 0x73: JUMP_DISP8(!test_cf(e)); break;
			case 0x74: JUMP_DISP8(test_zf(e)); break;
			case 0x75: JUMP_DISP8(!test_zf(e)); break;
			case 0x76: JUMP_DISP8(test_be(e)); break;
			case 0x77: JUMP_DISP8(!test_be(e)); break;
			case 0x78: JUMP_DISP8(test_sf(e)); break;
			case 0x79: JUMP_DISP8(!test_sf(e)); break;
			case 0x7a: JUMP_DISP8(test_pf(e)); break;
			case 0x7b: JUMP_DISP8(!test_pf(e)); break;
			case 0x7c: JUMP_DISP8(test_l(e)); break;	
			case 0x7d: JUMP_DISP8(!test_l(e)); break;
			case 0x7e: JUMP_DISP8(test_le(e)); break;
			case 0x7f: JUMP_DISP8(!test_le(e)); break;

			// Two-byte opcode
			case 0x0f:
				switch (*inp++) {

				// Jcc - conditional branch with disp32
				case 0x80: JUMP_DISP32(test_of(e), 0); break; 
				case 0x81: JUMP_DISP32(!test_of(e), 0); break;
				case 0x82: JUMP_DISP32(test_cf(e), 0); break;
				case 0x83: JUMP_DISP32(!test_cf(e), 0); break;
				case 0x84: JUMP_DISP32(test_zf(e), 0); break;
				case 0x85: JUMP_DISP32(!test_zf(e), 0); break;
				case 0x86: JUMP_DISP32(test_be(e), 0); break;
				case 0x87: JUMP_DISP32(!test_be(e), 0); break;
				case 0x88: JUMP_DISP32(test_sf(e), 0); break;
				case 0x89: JUMP_DISP32(!test_sf(e), 0); break;
				case 0x8a: JUMP_DISP32(test_pf(e), 0); break;
				case 0x8b: JUMP_DISP32(!test_pf(e), 0); break;
				case 0x8c: JUMP_DISP32(test_l(e), 0); break;
				case 0x8d: JUMP_DISP32(!test_l(e), 0); break;
				case 0x8e: JUMP_DISP32(test_le(e), 0); break;
				case 0x8f: JUMP_DISP32(!test_le(e), 0); break;
				default: goto invalid;
				}

			default: goto invalid;
			}
			break;

		case 0xc3: {					// RET
			register uint32_t target;
			POP(target);
printf("RET to %08x esp=%08x ebp=%08x\n", target, e->reg[ESP], e->reg[EBP]);
			JUMP(target);
			break;
		    }

		case 0xc2: {					// RET imm16
			register unsigned popcount = getu16l(inp); inp += 2;
			register uint32_t target;
			POP(target);
			e->reg[ESP] += popcount;
			JUMP(target);
			break;
		    }

		case 0xc8: {					// ENTER i16,i8
			unsigned framesize = getu16l(inp); inp += 2;
			unsigned nestlevel = *inp++;
			if (nestlevel != 0)
				goto invalid;	// XX ??
			PUSH(e->reg[EBP]);
			e->reg[EBP] = e->reg[ESP];
			e->reg[ESP] -= framesize;
			break;
		    }

		case 0xc9: {					// LEAVE
			uint32_t old_ebp;
			POP(old_ebp);
			e->reg[ESP] = e->reg[EBP];
			e->reg[EBP] = old_ebp;
			break;
		    }

		case 0xf8:					// CLC
			CHECKINP;
			update_eflags(e);
			e->eflags &= ~EFL_CF;
			break;
		case 0xf9:					// STC
			CHECKINP;
			update_eflags(e);
			e->eflags |= EFL_CF;
			break;
		case 0xf5:					// CMC
			CHECKINP;
			update_eflags(e);
			e->eflags ^= EFL_CF;
			break;


		// String instructions

		// XXX

		case 0xfc:					// CLD
			CHECKINP;
			e->eflags &= ~EFL_DF;
			break;
		case 0xfd:					// STD
			CHECKINP;
			e->eflags |= EFL_DF;
			break;


#if 0
		// INC reg; DEC reg; PUSH reg; POP reg; XCHG eAX,reg
		case 0x98: case 0x99:				// CWDE, CDQ
		case 0xa4: case 0xa5: case 0xa6: case 0xa7:	// MOVS, CMPS
		case 0xaa: case 0xab:				// STOS
		case 0xac: case 0xad: case 0xae: case 0xaf:	// LODS, SCAS
		case 0xfc: case 0xfd:				// CLD, STD
			goto notrans;

		// OP Eb,Ib; OP Ev,Ib; IMUL Gv,Ev,Ib
		case 0x80:					// OP Eb,Ib
		case 0x83:					// OP Ev,Ib
		case 0x6b:					// IMUL Gv,Ev,Ib
			inp = xscan_rm(inp);
			inp += 1;
			goto notrans;

		// OP Ev,Iv; IMUL Gv,Ev,Iv
		case 0x81:					// OP Ev,Iv
		case 0x69:					// IMUL Gv,Ev,Iv
			inp = xscan_rm(inp);
			inp += 4;
			goto notrans;

		// PUSHF; POPF
		case 0x9c: case 0x9d:
			assert(0);	// XXX

		case 0xcd:			// INT n (software interrupt)
			inp++;
		case 0xcc:			// INT3 (breakpoint)
			goto gentrap;

		// Group 4 - INC, DEC
		case 0xfe:
			ea = *inp;
			inp = xscan_rm(inp);
			switch (EA_REG(ea)) {
			case 0: case 1:			// INC Eb, DEC Eb
				goto notrans;
			}
			goto invalid;
#endif

		// Prefixes
		case 0x0f:	// 2-byte opcode escape
			goto twobyte;
		case 0x66:	// Operand size prefix
			goto opsize;
		case 0xf3:	// REP/REPE prefix
			goto rep;
		case 0xf2:	// REPNE prefix
			goto repne;

		default:
			goto invalid;
		}
		continue;

	// Operand size prefix (0x66) seen
	opsize:
		switch (*inp++) {

		// XXX

		// Prefixes
		case 0x0f:	// 2-byte opcode escape
			goto twobyte_opsize;
		case 0x66:	// Operand size prefix (redundant)
			goto invalid;
		case 0xf3:	// REP/REPE prefix
			goto opsize_rep;
		case 0xf2:	// REPNE prefix
			goto opsize_repne;

		default:
			goto invalid;
		}
		continue;

	// REP/REPE prefix (0xf3) seen
	rep:
		switch (*inp++) {

		// No-operand insns
		case 0xa4: case 0xa5: case 0xa6: case 0xa7:	// MOVS, CMPS
		case 0xaa: case 0xab:				// STOS
		case 0xac: case 0xad: case 0xae: case 0xaf:	// LODS, SCAS
			goto invalid;	// XXX

		// Prefixes
		case 0x0f:	// 2-byte opcode escape
			goto twobyte_rep;
		case 0x66:	// Operand size prefix
			goto opsize_rep;
		case 0xf3:	// REP/REPE prefix (redundant)
			goto invalid;
		case 0xf2:	// REPNE prefix (conflicting)
			goto invalid;

		default:
			goto invalid;
		}
		continue;

	// REPNE prefix (0xf2) seen
	repne:
		switch (*inp++) {

		// No-operand insns
		case 0xa6: case 0xa7:				// CMPS
		case 0xae: case 0xaf:				// SCAS
			goto invalid;	// XXX

		// Prefixes
		case 0x0f:	// 2-byte opcode escape
			goto twobyte_repne;
		case 0x66:	// Operand size prefix
			goto opsize_repne;
		case 0xf3:	// REP/REPE prefix (conflicting)
			goto invalid;
		case 0xf2:	// REPNE prefix (redundant)
			goto invalid;

		default:
			goto invalid;
		}
		continue;


	// Operand size prefix (0x66) and REP/REPE prefix (0xf3) seen
	opsize_rep:
		switch (*inp++) {
		case 0xa5: case 0xa7:				// MOVS, CMPS
		case 0xab:					// STOS
		case 0xad: case 0xaf:				// LODS, SCAS
			goto invalid;	// XXX

		default:
			goto invalid;
		}
		continue;

	// Operand size prefix (0x66) and REPNE prefix (0xf2) seen
	opsize_repne:
		switch (*inp++) {
		case 0xa7:					// CMPS
		case 0xaf:					// SCAS
			goto invalid;	// XXX

		default:
			goto invalid;
		}
		continue;


	twobyte:
		switch (*inp++) {

		// SYSCALL instruction for fast system calls
		case 0x05:
			e->eip = inp - e->mem;	// Save EIP _after_ insn
			e->trapno = VXTRAP_SYS;
			e->traperr = 0;
			e->trapva = 0;
			return VXTRAP_SYS;

		// Jcc rel32
		case 0x80: JUMP_DISP32(test_of(e), 0); break;	// JO rel32
		case 0x81: JUMP_DISP32(!test_of(e), 0); break;	// JNO rel32
		case 0x82: JUMP_DISP32(test_cf(e), 0); break;	// JC rel32
		case 0x83: JUMP_DISP32(!test_cf(e), 0); break;	// JNC rel32
		case 0x84: JUMP_DISP32(test_zf(e), 0); break;	// JZ rel32
		case 0x85: JUMP_DISP32(!test_zf(e), 0); break;	// JNZ rel32
		case 0x86: JUMP_DISP32(test_be(e), 0); break;	// JBE rel32
		case 0x87: JUMP_DISP32(!test_be(e), 0); break;	// JNBE rel32
		case 0x88: JUMP_DISP32(test_sf(e), 0); break;	// JS rel32
		case 0x89: JUMP_DISP32(!test_sf(e), 0); break;	// JNS rel32
		case 0x8a: JUMP_DISP32(test_pf(e), 0); break;	// JP rel32
		case 0x8b: JUMP_DISP32(!test_pf(e), 0); break;	// JNP rel32
		case 0x8c: JUMP_DISP32(test_l(e), 0); break;	// JL rel32
		case 0x8d: JUMP_DISP32(!test_l(e), 0); break;	// JNL rel32
		case 0x8e: JUMP_DISP32(test_le(e), 0); break;	// JLE rel32
		case 0x8f: JUMP_DISP32(!test_le(e), 0); break;	// JNLE rel32

		// SETcc
		case 0x90: SET_RM8(test_of(e)); break;		// SETO rm8
		case 0x91: SET_RM8(!test_of(e)); break;		// SETNO rm8
		case 0x92: SET_RM8(test_cf(e)); break;		// SETC rm8
		case 0x93: SET_RM8(!test_cf(e)); break;		// SETNC rm8
		case 0x94: SET_RM8(test_zf(e)); break;		// SETZ rm8
		case 0x95: SET_RM8(!test_zf(e)); break;		// SETNZ rm8
		case 0x96: SET_RM8(test_be(e)); break;		// SETBE rm8
		case 0x97: SET_RM8(!test_be(e)); break;		// SETNBE rm8
		case 0x98: SET_RM8(test_sf(e)); break;		// SETS rm8
		case 0x99: SET_RM8(!test_sf(e)); break;		// SETNS rm8
		case 0x9a: SET_RM8(test_pf(e)); break;		// SETP rm8
		case 0x9b: SET_RM8(!test_pf(e)); break;		// SETNP rm8
		case 0x9c: SET_RM8(test_l(e)); break;		// SETL rm8
		case 0x9d: SET_RM8(!test_l(e)); break;		// SETNL rm8
		case 0x9e: SET_RM8(test_le(e)); break;		// SETLE rm8
		case 0x9f: SET_RM8(!test_le(e)); break;		// SETNLE rm8

		// BSWAP
		case 0xc8: BSWAP_REG32(EAX); break;
		case 0xc9: BSWAP_REG32(ECX); break;
		case 0xca: BSWAP_REG32(EDX); break;
		case 0xcb: BSWAP_REG32(EBX); break;
		case 0xcc: BSWAP_REG32(ESP); break;
		case 0xcd: BSWAP_REG32(EBP); break;
		case 0xce: BSWAP_REG32(ESI); break;
		case 0xcf: BSWAP_REG32(EDI); break;

		// IMUL
		case 0xaf:				// IMUL r32,rm32
			ALU_REGRM32(e, dstarg * srcarg, STM_IMUL, 1);
			break;


#if 0
		// General EA operands
		case 0x10: case 0x11:			// MOVUPS
		case 0x12:				// MOVLPS Vps,Mq/MOVHLPS
		case 0x14: case 0x15:			// UNPCKLPS/UNPCKHPS
		case 0x16:				// MOVHPS Vps,Mq/MOVLHPS
		case 0x28: case 0x29:			// MOVAPS
		case 0x2e: case 0x2f:			// UCOMISS/COMISS
		case 0x40: case 0x41: case 0x42: case 0x43:	// CMOVcc
		case 0x44: case 0x45: case 0x46: case 0x47:
		case 0x48: case 0x49: case 0x4a: case 0x4b:
		case 0x4c: case 0x4d: case 0x4e: case 0x4f:
		case 0x51:					// SQRTPS
		case 0x54: case 0x55: case 0x56: case 0x57:	// ANDPS etc.
		case 0x58: case 0x59: case 0x5a: case 0x5b:	// ADDPS etc.
		case 0x5c: case 0x5d: case 0x5e: case 0x5f:	// SUBPS etc.
		case 0xa3:					// BT Ev,Gv
		case 0xab:					// BTS Ev,Gv
		case 0xb3:					// BTR Ev,Gv
		case 0xbb:					// BTC Ev,Gv
		case 0xbc: case 0xbd:				// BSF, BSR
		case 0xaf:					// IMUL Gv,Ev
		case 0xb6: case 0xb7:				// MOVZX
		case 0xbe: case 0xbf:				// MOVSX
			inp = xscan_rm(inp);
			goto notrans;

		// General EA operands plus immediate byte
		case 0xc2:				// CMPPS Vps,Wps,Ib
		case 0xc6:				// SHUFPS Vps,Wps,Ib
			inp = xscan_rm(inp);
			inp += 1;
			goto notrans;

		// Memory-only EA operand
		case 0x13:				// MOVLPS Mq,Vps
		case 0x17:				// MOVHPS Mq,Vps
		case 0x2b:				// MOVNTPS
		case 0xc3:				// MOVNTI Md,Gd
			if (EA_MOD(*inp) == 3)	// Mem-only
				goto invalid;
			inp = xscan_rm(inp);
			goto notrans;

		// Register-only EA operand
		case 0x50:				// MOVMSKPS
			if (EA_MOD(*inp) != 3)	// Reg-only
				goto invalid;
			inp = xscan_rm(inp);
			goto notrans;


		// Shift instructions
		case 0xa4:					// SHLD Ev,Gv,Ib
		case 0xac:					// SHRD Ev,Gv,Ib
			inp = xscan_rm(inp);
			inp += 1;
			// XXX fix cc
			goto notrans;
		case 0xa5:					// SHLD Ev,Gv,CL
		case 0xad:					// SHRD Ev,Gv,CL
			inp = xscan_rm(inp);
			// XXX fix cc
			goto notrans;

		// Group 8 - Bit test/modify with immediate
		case 0xba:
			if (!(EA_REG(*inp) & 4))
				goto invalid;
			inp = xscan_rm(inp);
			inp += 1;
			goto invalid;

		// Group 15 - SSE control
		case 0xae:
			ea = *inp;
			inp = xscan_rm(inp);
			switch (EA_REG(ea)) {
			case 2:					// LDMXCSR
			case 3:					// STMXCSR
				if (EA_MOD(ea) == 3)	// Mem-only
					goto invalid;
				goto notrans;
			// XX LFENCE, SFENCE, MFENCE?
			}
			goto invalid;

		// Group 16 - PREFETCH
		case 0x18:
			if (EA_MOD(*inp) == 3)	// Mem-only
				goto invalid;
			// XX Squash to NOP if EA_REG(*inp) > 3?
			inp = xscan_rm(inp);
			goto notrans;
#endif

		default:
			goto invalid;
		}
		continue;

	twobyte_opsize:
		switch (*inp++) {

#if 0
		// General EA operands
		case 0x10: case 0x11:			// MOVUPD
		case 0x14: case 0x15:			// UNPCKLPD/UNPCKHPD
		case 0x28: case 0x29:			// MOVAPD
		case 0x2e: case 0x2f:			// UCOMISD/COMISD
		case 0x40: case 0x41: case 0x42: case 0x43:	// CMOVcc
		case 0x44: case 0x45: case 0x46: case 0x47:
		case 0x48: case 0x49: case 0x4a: case 0x4b:
		case 0x4c: case 0x4d: case 0x4e: case 0x4f:
		case 0x51:					// SQRTPD
		case 0x54: case 0x55: case 0x56: case 0x57:	// ANDPD etc.
		case 0x58: case 0x59: case 0x5a: case 0x5b:	// ADDPD etc.
		case 0x5c: case 0x5d: case 0x5e: case 0x5f:	// SUBPD etc.
		case 0x60: case 0x61: case 0x62: case 0x63:	// PUNPCK...
		case 0x64: case 0x65: case 0x66: case 0x67:	// PCMPGT...
		case 0x68: case 0x69: case 0x6a: case 0x6b:	// PUNPCK...
		case 0x6c: case 0x6d: case 0x6e: case 0x6f:	// PUNPCK...
		case 0x74: case 0x75: case 0x76:		// PCMPEQ...
		case 0x7e: case 0x7f:				// MOVD/MOVDQA
		case 0xa3:					// BT Ev,Gv
		case 0xab:					// BTS Ev,Gv
		case 0xb3:					// BTR Ev,Gv
		case 0xbb:					// BTC Ev,Gv
		case 0xbc: case 0xbd:				// BSF, BSR
		case 0xaf:					// IMUL Gv,Ev
		case 0xb6:					// MOVZX Gv,Eb
		case 0xbe:					// MOVSX Gv,Eb
		case 0xd1: case 0xd2: case 0xd3:		// PSRLx
		case 0xd4: case 0xd5: case 0xd6:		// PADDQ...
		case 0xd8: case 0xd9: case 0xda: case 0xdb:	// PSUBUSB...
		case 0xdc: case 0xdd: case 0xde: case 0xdf:	// PADDUSB...
		case 0xe0: case 0xe1: case 0xe2: case 0xe3:	// PAVGB...
		case 0xe4: case 0xe5: case 0xe6:		// PMULHUW...
		case 0xe8: case 0xe9: case 0xea: case 0xeb:	// PSUBSB...
		case 0xec: case 0xed: case 0xee: case 0xef:	// PADDSB...
		case 0xf1: case 0xf2: case 0xf3:		// PSLLx
		case 0xf4: case 0xf5: case 0xf6:		// PMULUDQ...
		case 0xf8: case 0xf9: case 0xfa: case 0xfb:	// PSUBB...
		case 0xfc: case 0xfd: case 0xfe:		// PADDB...
			inp = xscan_rm(inp);
			goto notrans;

		// General EA operands plus immediate byte
		case 0xc5:				// PEXTRW Gd,VRdq,Ib
			if (EA_MOD(*inp) != 3)
				goto invalid; // Reg-only
		case 0x70:				// PSHUFD Vdq,Wdq,Ib
		case 0xc2:				// CMPPD Vps,Wps,Ib
		case 0xc4:				// PINSRW Vdq,Ew,Ib
		case 0xc6:				// SHUFPD Vps,Wps,Ib
			inp = xscan_rm(inp);
			inp += 1;
			goto notrans;

		// Memory-only EA operand
		case 0x12: case 0x13:			// MOVLPD
		case 0x16: case 0x17:			// MOVHPD
		case 0x2b:				// MOVNTPD
		case 0xe7:				// MOVNTDQ Mdq,Vdq
			if (EA_MOD(*inp) == 3)		// Mem-only
				goto invalid;
			inp = xscan_rm(inp);
			goto notrans;

		// Register-only EA operand
		case 0x50:				// MOVMSKPD
		case 0xd7:				// PMOVMSKB Gd,VRdq
		case 0xf7:				// MASKMOVQ Vdq,Wdq
			if (EA_MOD(*inp) != 3)		// Reg-only
				goto invalid;
			inp = xscan_rm(inp);
			goto notrans;

		// Shift instructions
		case 0xa4:					// SHLD Ev,Gv,Ib
		case 0xac:					// SHRD Ev,Gv,Ib
			inp = xscan_rm(inp);
			inp += 1;
			// XXX fix cc
			goto notrans;
		case 0xa5:					// SHLD Ev,Gv,CL
		case 0xad:					// SHRD Ev,Gv,CL
			inp = xscan_rm(inp);
			// XXX fix cc
			goto notrans;

		// Group 8 - Bit test/modify with immediate
		case 0xba:
			if (!(EA_REG(*inp) & 4))
				goto invalid;
			inp = xscan_rm(inp);
			inp += 1;
			goto invalid;

		// Group 12, 13, 14 - SSE vector shift w/ immediate
		case 0x71: case 0x72: case 0x73:
			ea = *inp;
			inp = xscan_rm(inp);
			switch (EA_REG(ea)) {
			case 2: case 4: case 6:
				inp += 1;
				goto notrans;
			}
			goto invalid;
#endif

		default:
			goto invalid;
		}
		continue;

	twobyte_rep:
		switch (*inp++) {

#if 0
		// General EA operands
		case 0x10: case 0x11:				// MOVSS
		case 0x2a: case 0x2c: case 0x2d:		// CVT...
		case 0x51:					// SQRTSS
		case 0x58: case 0x59: case 0x5a: case 0x5b:	// ADDSS etc.
		case 0x5c: case 0x5d: case 0x5e: case 0x5f:	// SUBSS etc.
		case 0x6f:					// MOVDQU
		case 0x7e: case 0x7f:				// MOVQ/MOVDQU
		case 0xe6:					// CVTDQ2PD
			inp = xscan_rm(inp);
			goto notrans;

		// General EA operands plus immediate byte
		case 0x70:				// PSHUFHW Vq,Wq,Ib
		case 0xc2:				// CMPSS Vss,Wss,Ib
			inp = xscan_rm(inp);
			inp += 1;
			goto notrans;
#endif

		default:
			goto invalid;
		}
		continue;

	twobyte_repne:
		switch (*inp++) {

#if 0
		// General EA operands
		case 0x10: case 0x11:				// MOVSD
		case 0x2a: case 0x2c: case 0x2d:		// CVT...
		case 0x51:					// SQRTSD
		case 0x58: case 0x59: case 0x5a:		// ADDSD etc.
		case 0x5c: case 0x5d: case 0x5e: case 0x5f:	// SUBSD etc.
		case 0xe6:					// CVTPD2DQ
			inp = xscan_rm(inp);
			goto notrans;

		// General EA operands plus immediate byte
		case 0x70:				// PSHUFLW Vq,Wq,Ib
		case 0xc2:				// CMPSD Vss,Wss,Ib
			inp = xscan_rm(inp);
			inp += 1;
			goto notrans;
#endif

		default:
			goto invalid;
		}
		continue;


	invalid:
		fprintf(stderr, "invalid opcode %02x %02x %02x at eip %08x\n",
			e->ininst[0], e->ininst[1], e->ininst[2],
			e->ininst - e->mem);

		// Instruction wasn't executed, so save EIP before insn
		e->eip = e->ininst - e->mem;
		e->trapno = VXTRAP_INVALID;
		e->traperr = 0;
		e->trapva = 0;
		return VXTRAP_INVALID;


	diverr:
		e->eip = e->ininst - e->mem;
		e->trapno = VXTRAP_DIVIDE;
		e->traperr = 0;
		e->trapva = 0;
		return VXTRAP_DIVIDE;


	noexec:
		// We overshot inmax while decoding an instruction.
		// If we successfully executed at least one instruction,
		// then we might have just crossed a page boundary,
		// so go back and recheck instruction permissions.
		e->eip = e->ininst - e->mem;
		if (e->ininst != instart)
			goto neweip;

		// Part of this instruction is not executable, so trap.
		e->trapno = VXTRAP_PAGEFAULT;
		e->traperr = 0x10;		// Execute fault
		e->trapva = e->eip;
		return VXTRAP_PAGEFAULT;

	} // while(1)

	// Clear the special instruction-interpretation exception state flag
	e->ininst = NULL;

	return 0;
}


void vxproc_setpcall(vxproc *vxp, vxpcallhandler *handler, void *pcalldata,
			size_t maxmsglen)
{
	vxemu *emu = vxp->emu;

	emu->pcallhandler = handler;
	emu->pcalldata = pcalldata;
	emu->maxmsglen = maxmsglen;
}

// Handle system calls in the VX environment
static int dosyscall(struct vxproc *vxp)
{
	vxemu *emu = vxp->emu;

	// SYSENTER always copies the caller's EIP into ECX.
	emu->reg[ECX] = emu->eip;

	// Check the system call code in EAX.
	uint32_t code = emu->reg[EAX];
	switch (code & 0xff) {

	case VXSYS_PCALL:	// Parent-call
		if (emu->pcallhandler != NULL) {

			vxpmsg m;
			m.code = code;
			m.val1 = emu->reg[ESI];
			m.val2 = emu->reg[EDI];

			// Validate the input message if needed
			if (code & VXPC_SENDBUF) {
				uint32_t va = emu->reg[EBX];
				uint32_t len = emu->reg[EDX];
				size_t faultva;
				if (!vxproc_checkperm(vxp, va, len,
						VXPERM_READ, &faultva)) {
					emu->traperr = 4;	// read fault
					emu->trapva = faultva;
					return (emu->trapno = VXTRAP_PAGEFAULT);
				}
				m.data = vxp->data + va;
				m.datalen = len;
			} else {
				m.valB = emu->reg[EBX];
				m.valL = emu->reg[EDX];
			}

			// Call the registered handler
			int rc = emu->pcallhandler(vxp, &m, emu->pcalldata);
			if (rc != 0)
				return rc;

			// XX send reply message if needed

			emu->reg[EAX] = m.code;
			emu->reg[ESI] = m.val1;
			emu->reg[EDI] = m.val2;

			// Copy ECX back to EIP as per SYSEXIT behavior,
			// in case pcallhandler changed ECX.
			emu->eip = emu->reg[ECX];
			return 0;
		}
		break;

	case VXSYS_SETPERM:	// Set permissions
		if (vxproc_setperm(vxp, emu->reg[EBX], emu->reg[EDX],
				emu->reg[ESI]) < 0) {
			if (errno == EINVAL) {
				// Invalid system call argument - trap.
				return (emu->trapno = VXTRAP_INVARG);
			}

			// Otherwise, it's a host thing such as no memory -
			// pass it back to the host client.
			return -1;
		}
		return 0;
		break;

	}

	// Default: produce a VXTRAP_SYS trap
	return (emu->trapno = VXTRAP_SYS);
}

int vxproc_run(struct vxproc *vxp)
{
	vxemu *emu = vxp->emu;

	// Make sure the process is mapped into our host memory
	if (vxproc_map(vxp) < 0)
		return -1;

	// Save our stack environment for exception-handling
	jmp_buf env;
	int rc = setjmp(env);
	if (rc) {
		fprintf(stderr, "VX trap %x err %x va %08x "
				"veip %08x veflags %08x\n",
				emu->trapno, emu->traperr, emu->trapva,
				emu->eip, emu->eflags);
		return rc;
	}
	emu->trapenv = &env;

	while (1) {

		// Interpret instructions until we hit an exception
		rc = emulate(emu);
		if (rc != VXTRAP_SYS)
			break;

		// Handle system call.
		rc = dosyscall(vxp);
		if (rc != 0)
			break;

		rc = vxproc_map(vxp);
		if (rc < 0)
			break;
	}

	// De-register our setjmp environment for trap handling.
	emu->trapenv = NULL;

	return rc;
}

void vxemu_flush(struct vxemu *e)
{
	// Nothing to do, since we don't cache instructions.
}

int vxemu_map(struct vxemu *e)
{
	// Keep mirror copies of the memory pointer and size
	e->mem = e->proc->data;
	e->memsize = e->proc->size;

	return 0;
}

