/*
 * Slow but highly portable machine-word access primitives.
 */
#ifndef INTERP_WORDS_H
#define INTERP_WORDS_H

#include "config.h"


// System-independent but still moderately efficient byte-swapping functions
static inline uint16_t bswap16(uint16_t v)
{
	return (v >> 8) | (v << 8);
}

static inline uint32_t bswap32(uint32_t v)
{
	uint32_t t = ((v & 0xff00ff00) >> 8) | ((v & 0x00ff00ff) << 8);
	return (t >> 16) | (t << 16);
}

static inline uint64_t bswap64(uint64_t v)
{
	uint64_t b =	((v & 0xff00ff00ff00ff00LL) >> 8) |
			((v & 0x00ff00ff00ff00ffLL) << 8);
	uint64_t w =	((b & 0xffff0000ffff0000LL) >> 16) |
			((b & 0x0000ffff0000ffffLL) << 16);
	return (w >> 32) | (w << 32);
}


// Swap between host and big or little-endian byte orders
#ifdef WORDS_BIGENDIAN
#define htol16(x)	bswap16(x)
#define ltoh16(x)	bswap16(x)
#define htol32(x)	bswap32(x)
#define ltoh32(x)	bswap32(x)
#define htol64(x)	bswap64(x)
#define ltoh64(x)	bswap64(x)
#define htob16(x)	(x)
#define btoh16(x)	(x)
#define htob32(x)	(x)
#define btoh32(x)	(x)
#define htob64(x)	(x)
#define btoh64(x)	(x)
#else
#define htol16(x)	(x)
#define ltoh16(x)	(x)
#define htol32(x)	(x)
#define ltoh32(x)	(x)
#define htol64(x)	(x)
#define ltoh64(x)	(x)
#define htob16(x)	bswap16(x)
#define btoh16(x)	bswap16(x)
#define htob32(x)	bswap32(x)
#define btoh32(x)	bswap32(x)
#define htob64(x)	bswap64(x)
#define btoh64(x)	bswap64(x)
#endif



// Read words from potentially-unaligned memory locations
// in little-endian byte order.

static inline uint16_t getu16l(const uint8_t *inp)
{
	return	((uint16_t)inp[0]) |
		((uint16_t)inp[1] << 8);
}

static inline uint32_t getu32l(const uint8_t *inp)
{
	return	((uint32_t)inp[0]) |
		((uint32_t)inp[1] << 8) |
		((uint32_t)inp[2] << 16) |
		((uint32_t)inp[3] << 24);
}

static inline uint64_t getu64l(const uint8_t *inp)
{
	return	((uint64_t)inp[0]) |
		((uint64_t)inp[1] << 8) |
		((uint64_t)inp[2] << 16) |
		((uint64_t)inp[3] << 24) |
		((uint64_t)inp[4] << 32) |
		((uint64_t)inp[5] << 40) |
		((uint64_t)inp[6] << 48) |
		((uint64_t)inp[7] << 56);
}


// Read words from potentially-unaligned memory locations
// in network (big-endian) byte order.

static inline uint16_t getu16n(const uint8_t *inp)
{
	return	(inp[0] << 8) |
		(inp[1]);
}

static inline uint32_t getu32n(const uint8_t *inp)
{
	return	((uint32_t)inp[0] << 24) |
		((uint32_t)inp[1] << 16) |
		((uint32_t)inp[2] << 8) |
		((uint32_t)inp[3]);
}

static inline uint64_t getu64n(const uint8_t *inp)
{
	return	((uint64_t)inp[0] << 56) |
		((uint64_t)inp[1] << 48) |
		((uint64_t)inp[2] << 40) |
		((uint64_t)inp[3] << 32) |
		((uint64_t)inp[4] << 24) |
		((uint64_t)inp[5] << 16) |
		((uint64_t)inp[6] << 8) |
		((uint64_t)inp[7]);
}


// Read words from potentially-unaligned memory locations
// in host byte order, whichever that is.
#ifdef WORDS_BIGENDIAN
#define getu16(p) getu16n(p)
#define getu32(p) getu32n(p)
#define getu64(p) getu64n(p)
#else
#define getu16(p) getu16l(p)
#define getu32(p) getu32l(p)
#define getu64(p) getu64l(p)
#endif


// Write words to potentially-unaligned memory locations
// in little-endian byte order.

static inline void putu16l(uint8_t *inp, uint16_t val)
{
	inp[0] = val;
	inp[1] = val >> 8;
}

static inline void putu32l(uint8_t *inp, uint32_t val)
{
	inp[0] = val;
	inp[1] = val >> 8;
	inp[2] = val >> 16;
	inp[3] = val >> 24;
}

static inline void putu64l(uint8_t *inp, uint64_t val)
{
	inp[0] = val;
	inp[1] = val >> 8;
	inp[2] = val >> 16;
	inp[3] = val >> 24;
	inp[4] = val >> 32;
	inp[5] = val >> 40;
	inp[6] = val >> 48;
	inp[7] = val >> 56;
}


// Write words to potentially-unaligned memory locations
// in network (big-endian) byte order.

static inline void putu16n(uint8_t *inp, uint16_t val)
{
	inp[0] = val >> 8;
	inp[1] = val;
}

static inline void putu32n(uint8_t *inp, uint32_t val)
{
	inp[0] = val >> 24;
	inp[1] = val >> 16;
	inp[2] = val >> 8;
	inp[3] = val;
}

static inline void putu64n(uint8_t *inp, uint64_t val)
{
	inp[0] = val >> 56;
	inp[1] = val >> 48;
	inp[2] = val >> 40;
	inp[3] = val >> 32;
	inp[4] = val >> 24;
	inp[5] = val >> 16;
	inp[6] = val >> 8;
	inp[7] = val;
}


// Write words to potentially-unaligned memory locations
// in host byte order, whichever that is.
#ifdef WORDS_BIGENDIAN
#define putu16(p, v) putu16n(p, v)
#define putu32(p, v) putu32n(p, v)
#define putu64(p, v) putu64n(p, v)
#else
#define putu16(p, v) putu16l(p, v)
#define putu32(p, v) putu32l(p, v)
#define putu64(p, v) putu64l(p, v)
#endif


#endif	// INTERP_WORDS_H
