//
// Main private header file for vx32-to-PowerPC instruction translator.
//
#ifndef EMU_PPC_H
#define EMU_PPC_H

#include <inttypes.h>
#include <setjmp.h>

#include <vx/proc.h>			// for vxpcallhandler


struct vxproc;


// Map of how PowerPC registers are used by translated code.
#define PR_VXREG	8		// Eight x86 general-purpose registers
#define PR_CCMODE	16		// Condition codes mode
#define PR_CCRES	17		// Condition codes result value
#define PR_CCARG	18		// Condition codes argument value
#define PR_MEMBASE	19		// Base of vx32 memory region
#define PR_MEMSIZE	20		// Current size of vx32 region
#define PR_LINKEIP	21		// vx32 EIP for current PPC Link Reg
#define PR_AMSK8	22		// Address mask for 8-bit operands
#define PR_AMSK16	23		// Address mask for 16-bit operands
#define PR_AMSK32	24		// Address mask for 32-bit operands


typedef struct vxinsn {
	uint8_t		itype;
	uint8_t		srcofs;		// Offset in original vx32 code
	uint8_t		dstofs;		// Offset in translated x86 code
	uint8_t		dstlen;		// Dest x86 code size for this insn
} vxinsn;
// Instruction types for vxinsn.itype
#define VXI_NOTRANS	0x00		// Simple, no translation required
#define VXI_JUMP	0x02		// Direct jump/branch/call instruction
#define VXI_JUMP8	0x03		// 8-bit jump/call/branch instruction
#define VXI_INDIR	0x04		// Indirect jump/call/ret
#define VXI_TRAP	0x05		// Trapping instruction
#define VXI_ENDFRAG	0x06		// Terminate this fragment


// Limits for vxinsn.srcofs and vxinsn.dstofs
#define VXSRCOFS_MAX	255
#define VXDSTOFS_MAX	255



// Each translated fragment consists of:
//	- A struct vxfrag header
//	- A translation summary table containing ninsn vxinsn structs.
//	- The translated code fragment itself, as laid out in the table.
//	- Any trampoline code required by jumps within the fragment.
typedef struct vxfrag {
	uint32_t	eip;		// Original EIP in emulated vx32 code
	uint8_t		ninsn;		// Number of translated vx32 insns
	struct vxinsn	insn[0];	// Beginning of vxinsn table
} vxfrag;

// Macro to find the beginning of translated code in a fragment
#define FRAGCODE(f)	((uint8_t*)&(f)->insn[(f)->ninsn])


// The vx32-EIP-to-translated-entrypoint hash table
// consists of vxentry structures.
// Each vxentry maps a EIP value in the original vx32 code
// to a corresponding PC in translated code.
// Empty entries are indicated by *dstip* being 0.
// (Zero is theoretically, though unlikely in practice, a valid vx32 EIP.)
typedef struct vxentry {
	uint32_t	srceip;
	uint32_t	dsteip;
} vxentry;

// Emulation state for vx32-to-x86 translation.
// This is the header for a variable-length structure;
// the variable-length part is the vx32 EIP lookup hash table (etab),
// followed by the instruction translation buffer.
// While running a particular vxproc,
// we always keep the %ds,%es,%ss segment registers loaded with a data segment
// that allows access only to the emulated address space region, and
// we always keep the %fs register loaded with a special data segment
// whose base and limit correspond to the address and size of this state area.
// This way the emulation code can always access the emulation state
// at fixed displacements without having to use any general-purpose registers.
struct vxemu {

	// Back pointer to vxproc
	struct vxproc	*proc;

	// vx32 virtual CPU state
	uint32_t	reg[8];
	uint32_t	eip;
	uint32_t	eflags;

	// Virtual trap handling state
	uint32_t	trapno;
	uint32_t	traperr;
	uint32_t	trapva;
	jmp_buf		*trapenv;

	// General emulation state
	vxpcallhandler	*pcallhandler;	// Registered parent-call handler
	void		*pcalldata;	// Opaque data assoc'd with handler
	size_t		maxmsglen;	// Maximum data to recv on pcalls	
	uint32_t	emuflags;	// Emulation flags (see below)
	uint32_t	jmpinfo;	// for jump insn backpatching

	// Max # of instructions to execute before returning
	uint32_t	insncount;

	// Instruction scanning/translation state
	struct vxfrag	*txfrag;	// frag currently under construction
	uint8_t		*ininst;	// current instruction being scanned
	uint8_t		*instop;	// stop translating instructions here
	uint8_t		dstlen;		// # dest ppc insns for this insn
	uint32_t	insdisp;	// displacement in current insn's EA

	// Translated code buffer.
	// We write code frags into it from the bottom up, increasing codefree.
	// Each frag consists of a vxfrag, a translation summary, and the code.
	// We also keep a table of offsets of all the vxfrag headers,
	// starting at the top of the codebuf and working down, via codetab.
	// The frag header offsets are thus stored in opposite order
	// as the target code fragments themselves,
	// and the codefree and codefrags pointers work towared each other
	// until they collide, at which point the translation buffer is reset.
	void		*codebuf;
	void		*codefree;
	void		*codetab;
	void		*codetop;

	// Hash table for looking up entrypoints in original code
	uint32_t	etablen;	// Number of entries (power of two)
	uint32_t	etabmask;	// etablen-1
	uint32_t	etabcnt;	// Number of entries currently in use
	struct vxentry	etab[0];	// Must be last!
};
typedef struct vxemu vxemu;

// Hash function for entrypoint hash table.
// Keep it consistent with assembly code in x86asm.S!
#define etabhash(va)	(((((va) >> 10) + (va)) >> 10) - (va))

// Emulation flags (emuflags)
#define EMU_FP		0x01	// Enable SSE2 scalar floating-point
#define EMU_VFP		0x02	// Enable vector floating-point
#define EMU_VINT	0x04	// Enable vector integer arithmetic


// run.S
void vxrun_setup(vxemu *emu);
void vxrun_cleanup(vxemu *emu);
int vxrun(vxemu *emu, uint32_t dstip);


#endif	// EMU_PPC_H
