
// Benchmark program to test different (strong and weak) hash algorithms.
// Results using gcc -O9 as of 5/31/05:
//
//	adler32:  656745733 BPS
//	xmm32:    632892323 BPS
//	SHA-1:    100991796 BPS
//	SHA512:    28140372 BPS
//
// So unfortunately it appears trying to SIMD-optimize adler32
// probably isn't very fruitful, unless I've missed major optimizations.
// The problem is probably that the basic adler32 is fast enough
// so it's entirely memory-bound rather than CPU-bound.



#include <stdio.h>
#include <inttypes.h>
#include <sys/time.h>

#include <openssl/sha.h>

extern "C" {
#include "sha2.h"

uint32_t xmmcsum(uint32_t initcsum, void *buf, size_t nblocks);
}

#define min(x,y) ((x) < (y) ? (x) : (y))


typedef int64_t uipdate;

static inline uipdate tv2date(const struct timeval *tv)
{
	return (uipdate)tv->tv_sec * 1000000 + tv->tv_usec;
}

static inline uipdate getdate()
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return tv2date(&tv);
}

// Compute an Adler-32 checksum over a given block of data.
uint32_t sumblock(const void *buf, int len)
{
	const uint8_t *bf = (const uint8_t*)buf;
	unsigned a = 0;
	unsigned b = 0;
	for (int j = 0; j < len; ) {

		// We can sum up to 5552 bytes at a time (RFC 1950)
		// before our unsigned 32-bit accumulators overflow.
		for (int k = min(j+5550, len); j < k; j++) {
			a += *bf++;
			b += a;
		}
		a %= 65521;
		b %= 65521;
	}
	return (b << 16) | a;
}


#define REPS	1000

uint8_t buf[65536];

void SHA512(const uint8_t* buf, size_t sz, uint8_t dig[SHA512_DIGEST_LENGTH])
{
	SHA512_CTX ctx;
	SHA512_Init(&ctx);
	SHA512_Update(&ctx, buf, sz);
	SHA512_Final(dig, &ctx);
}

union xmmreg {
	uint8_t __attribute__ ((aligned(16))) b[16];
	uint16_t s[8];
	uint32_t w[4];
	uint64_t l[2];
};

int main(int argc, char **argv)
{
	uipdate t1, t2;

	t1 = getdate();
	for (int i = 0; i < REPS; i++) {
		sumblock(buf, sizeof(buf));
	}
	t2 = getdate();
	printf("adler32: %10.0f BPS\n", (float)REPS*sizeof(buf)*1000000.0/(t2-t1));

	t1 = getdate();
	for (int i = 0; i < REPS; i++) {
		xmmcsum(0, buf, sizeof(buf)/16);
	}
	t2 = getdate();
	printf("xmm32:   %10.0f BPS\n", (float)REPS*sizeof(buf)*1000000.0/(t2-t1));

	t1 = getdate();
	for (int i = 0; i < REPS; i++) {
		SHA_CTX ctx;
		uint8_t dig[SHA_DIGEST_LENGTH];
		SHA1_Init(&ctx);
		SHA1_Update(&ctx, buf, sizeof(buf));
		SHA1_Final(dig, &ctx);
	}
	t2 = getdate();
	printf("SHA-1:   %10.0f BPS\n", (float)REPS*sizeof(buf)*1000000.0/(t2-t1));

	t1 = getdate();
	for (int i = 0; i < REPS; i++) {
		uint8_t dig[SHA512_DIGEST_LENGTH];
		SHA512(buf, sizeof(buf), dig);
	}
	t2 = getdate();
	printf("SHA512:  %10.0f BPS\n", (float)REPS*sizeof(buf)*1000000.0/(t2-t1));
}

