
// Results as of 5/31/05 on my Opteron 240:
//	
//	Trivial test pattern:
//	naive by bytes : 23538.804688 cycles
//	naive by words : 5728.751953 cycles
//	xmm-based: 3193.653076 cycles
//	xmm, unrolled: 3071.821045 cycles
//	
//	Byte-interleaved:
//	naive by bytes : 21666.960938 cycles
//	xmm-based: 3180.645020 cycles
//	xmm, unrolled: 3007.300049 cycles
//	
//	Word-interleaved:
//	naive by bytes : 27529.935547 cycles
//	naive by words : 5372.613770 cycles
//	xmm-based: 3243.960938 cycles
//	xmm, unrolled: 3005.836914 cycles
//
// Hooray for SIMD!


#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <inttypes.h>

static inline long rdtsc()
{
	unsigned a, d;
	asm volatile("rdtsc" : "=d" (d), "=a" (a));
	return ((long)d << 32) | a;
}

extern void rcdifftest();

#define REPS 1000


uint8_t testpages[3][4096] __attribute__ ((aligned (16)));

int rcdbytes(const uint8_t *orig, const uint8_t *new, uint8_t *accum)
{
	int conflict = 0;
	for (int i = 0; i < 4096; i++) {
		uint8_t ob = *orig;
		uint8_t nb = *new;
		if (nb != ob) {
			if (*accum != ob)
				conflict = 1;
			*accum = nb;
		}
		orig++, new++, accum++;
	}
	return conflict;
}

int rcdints(const uint32_t *orig, const uint32_t *new, uint32_t *accum)
{
	int conflict = 0;
	for (int i = 0; i < 1024; i++) {
		uint32_t ob = *orig;
		uint32_t nb = *new;
		if (nb != ob) {
			if (*accum != ob)
				conflict = 1;
			*accum = nb;
		}
		orig++, new++, accum++;
	}
	return conflict;
}

// XMM-optimized diff/merge routines
extern int rcdiff(const void *orig, const void *new, void *accum);
extern int rcdiffu(const void *orig, const void *new, void *accum);


void runtests(int granularity)
{
	long t1, t2;
	int cfl = 0;
	int i;

	sleep(1);
	t1 = rdtsc();
	for (i = 0; i < REPS; i += 2) {
		cfl |= rcdbytes(testpages[0], testpages[1], testpages[2]);
		cfl |= rcdbytes(testpages[1], testpages[0], testpages[2]);
	}
	t2 = rdtsc();
	printf("naive by bytes : %f cycles\n", (float)(t2-t1) / REPS);
	if (cfl) printf("Conflicts!\n");

	if (granularity >= 4) {
		sleep(1);
		t1 = rdtsc();
		for (i = 0; i < REPS; i += 2) {
			cfl |= rcdints(
				(const uint32_t*)testpages[0],
				(const uint32_t*)testpages[1],
				(uint32_t*)testpages[2]);
			cfl |= rcdints(
				(const uint32_t*)testpages[1],
				(const uint32_t*)testpages[0],
				(uint32_t*)testpages[2]);
		}
		t2 = rdtsc();
		printf("naive by words : %f cycles\n", (float)(t2-t1) / REPS);
		if (cfl) printf("Conflicts!\n");
	}

	sleep(1);
	t1 = rdtsc();
	for (i = 0; i < REPS; i += 2) {
		cfl |= rcdiff(testpages[0], testpages[1], testpages[2]);
		cfl |= rcdiff(testpages[1], testpages[0], testpages[2]);
	}
	t2 = rdtsc();
	printf("xmm-based: %f cycles\n", (float)(t2-t1) / REPS);
	if (cfl) printf("Conflicts!\n");

	sleep(1);
	t1 = rdtsc();
	for (i = 0; i < REPS; i += 2) {
		cfl |= rcdiffu(testpages[0], testpages[1], testpages[2]);
		cfl |= rcdiffu(testpages[1], testpages[0], testpages[2]);
	}
	t2 = rdtsc();
	printf("xmm, unrolled: %f cycles\n", (float)(t2-t1) / REPS);
	if (cfl) printf("Conflicts!\n");
}

int main()
{
	int i;

	// Trivial test pattern: no changes
	printf("\nTrivial test pattern:\n");
	for (i = 0; i < 4096; i++) {
		testpages[0][i] = testpages[1][i] = testpages[2][i] = i;
	}
	runtests(4);

	// More difficult test pattern: non-conflicting,
	// but many changes to merge.
	printf("\nByte-interleaved:\n");
	for (i = 0; i < 4096; i++) {
		unsigned seq = (i % 3);
		testpages[0][i] = i;
		testpages[1][i] = i + (seq == 1);
		testpages[2][i] = i + (seq == 2);
	}
	runtests(1);

	printf("\nWord-interleaved:\n");
	for (i = 0; i < 4096; i++) {
		unsigned seq = ((i/4) % 3);
		testpages[0][i] = i;
		testpages[1][i] = i + (seq == 1);
		testpages[2][i] = i + (seq == 2);
	}
	runtests(4);
}

