
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>

#include <vx/proc.h>
#include <vx/vx32.h>


const char *progname;

void fatal(const char *fmt, ...)
{
	va_list ap;
	fprintf(stderr, "%s: fatal error: ", progname);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fputc('\n', stderr);
	exit(2);
}

void vxemu_stats(struct vxproc *p);	// XXX

void dumpregs(struct vxproc *p)
{
	struct vx32regs r;
	vx32_getregs(p, &r);

	fprintf(stderr, "eax %08x  ecx %08x  edx %08x  ebx %08x\n",
		r.eax, r.ecx, r.edx, r.ebx);
	fprintf(stderr, "esp %08x  ebp %08x  esi %08x  ebx %0di\n",
		r.esp, r.ebp, r.esi, r.edi);
	fprintf(stderr, "eip %08x  eflags %08x\n",
		r.eip, r.eflags);

	for (int i = 0; i < 8; i++) {
		int32_t *val = r.xmm[i].i32;
		fprintf(stderr, "xmm%d %08x%08x%08x%08x\n",
			i, val[3], val[2], val[1], val[0]);
	}
}

int pcallhandler(vxproc *proc, vxpmsg *msg, void *pcalldata)
{
	ssize_t rc = 0;

	switch (msg->code) {

	case VXPC_EXIT:
		//vxemu_stats(proc);
		exit(msg->val1);

	case VXPC_WRITE:
		if (msg->val1 > 2)
			fatal("bad file handle %d to write()", msg->val1);
		rc = write(msg->val1, msg->data, msg->datalen);
		break;

	case VXPC_READ:
		if (msg->val1 != 0)
			fatal("bad file handle %d to read()", msg->val1);
		void *p = vxproc_maprange(proc, msg->valB, msg->valL,
						VXPERM_WRITE);
		if (p == NULL)
			fatal("bad write buffer %08x size %08x\n",
				msg->valB, msg->valL);
		rc = read(msg->val1, p, msg->valL);
		vxproc_unmaprange(proc);
		break;

	default:
		dumpregs(proc);
		fatal("vxrun: invalid parent-call %x", msg->code);
	}

	if (rc < 0) {
		msg->code = -errno;
		return 0;
	}
	msg->code = rc;
	return 0;
}

int main(int argc, char **argv)
{
	progname = argv[0];

	if (argc < 2) {
		fprintf(stderr, "Usage: %s <vx-program> <args>\n",
			progname);
		exit(1);
	}
	const char *loadname = argv[1];

	vxproc *p = vxproc_new();
	if (p == NULL)
		fatal("vxproc_new: %s\n", strerror(errno));

	if (vxproc_loadfile(p, loadname, &argv[1], NULL) < 0)
		fatal("vxproc_loadfile: %s\n", strerror(errno));

	if (vxproc_siginit() < 0)
		fatal("vxproc_siginit: %s\n", strerror(errno));

	vxproc_setpcall(p, pcallhandler, NULL, (size_t)-1);

	int rc = vxproc_run(p);
	if (rc < 0)
		fatal("vxproc_run: %s\n", strerror(errno));
	fprintf(stderr, "vxproc_run returned %x\n", rc);

	vxproc_free(p);

	return 0;
}

