//
// Interprocess communication (IPC) in the VX environment
#ifndef VX_IPC_H
#define VX_IPC_H

#include <stddef.h>
#include <inttypes.h>


// IPC control word layout
#define VXIPC_HANDLE		0xffc00000	// Handle to invoke
#define VXIPC_SENDTYPE		0x00300000	// Send argument type
#define VXIPC_RECVTYPE		0x000c0000	// Receive argument type
#define VXIPC_IPCTYPE		0x00030000	// Basic IPC function
#define VXIPC_REQCODE		0x0000ffff	// Request/reply code

#define VXIPC_SEND0		0x00000000	// Send two simple values
#define VXIPC_SEND1		0x00100000	// Send one handle, one value
#define VXIPC_SEND2		0x00200000	// Send two handles
#define VXIPC_SENDBUF		0x00300000	// Send contents of data buffer

#define VXIPC_RECV0		0x00000000	// Receive two simple values
#define VXIPC_RECV1		0x00040000	// Receive one handle, one value
#define VXIPC_RECV2		0x00080000	// Receive two handles
#define VXIPC_RECVBUF		0x000c0000	// Receive into data buffer

#define VXIPC_CALL		0x00000000	// Invoke call handle
#define VXIPC_RETURN		0x00010000	// Return from call



// Request code assignments.
// By convention, the 16-bit request code is logically broken
// into three components for assignment purposes:
//	- 8-bit subsystem ID
//	- 4-bit interface ID
//	- 4-bit request ID
#define VXIPC_SUBSYS		0x0000ff00	// Subsystem ID
#define VXIPC_INTF		0x000000f0	// Interface ID
#define VXIPC_REQ		0x0000000f	// Request ID

// Current subsystem ID assignments
#define VXSYS_VX		0x00000000	// Basic VX environment
#define VXSYS_VXA		0x00000100	// VXA archiver
#define VXSYS_VXDOC		0x00000200	// VXDOC active document system

// Current interface ID assignments for VX subsystem
#define VXIF_PROC		0x00000010	// Process control interface
#define VXIF_CHUNK		0x00000020	// Memory chunk interface
#define VXIF_FILE		0x00000030	// File I/O interface


// Special "request" code used only for error replies.
// Not logically part of any interface.
#define VXERRCODE	0x0000000f	// Error: specific code is in value 1


// Process control request codes.
// Processes normally keep their own process control handle as handle 3,
// but parent environments can interpose on a child's self-handle if desired.
#define VXPROC_EXIT	0x00000010	// Terminate process with a status code
#define VXPROC_DROP	0x00000011	// Drop a key (or release a chunk)
#define VXPROC_GET	0x00000012	// Extract a handle from process
#define VXPROC_PUT	0x00000013	// Insert a handle into process
#define VXPROC_ALLOC	0x00000014	// Allocate a new zero-filled chunk
#define VXPROC_MKCALL	0x00000015	// Create call handle for this process


// File I/O interface request codes
#define VXFILE_READ	0x00000030	// Read stream
#define VXFILE_WRITE	0x00000031	// Write stream
#define VXFILE_SEEK	0x00000032	// Seek in stream



// A VX key is represented by an opaque void* pointer
// in this high-level interface.
// Within a VX environment, the low 22 bits are insignificant
// for purposes of selecting the key in the process's key table.
// In a host OS, however, the lower bits may be significant,
// so code designed to run either in a VX process or a host process
// must treat the entire pointer as opaque.
typedef void vxkey;


// CALL argument struct used with the high-level vx_call() function.
union vxargs {
	struct {			// Simple register values
		intptr_t	v1;	// Value 1 (VXIPC_SEND0)
		intptr_t	v2;	// Value 2 (VXIPC_SEND0/SEND1)
	};
	struct {			// Keys to transfer
		vxkey		*k1;	// Key 1 (VXIPC_SEND1/SEND2)
		vxkey		*k2;	// Key 2 (VXIPC_SEND2)
		int		o1;	// 22-bit offset 1 (SEND1/SEND2)
		int		o2;	// 22-bit offset 2 (SEND2)
	};
	struct {			// Data buffer to copy (VXIPC_SENDBUF)
		const void	*buf;	// Buffer pointer
		size_t		len;	// Number of bytes to send
	};
};
struct vxcall {
	union vxargs	send;			// Arguments to send
	union vxargs	recv;			// Arguments to receive
};



// Invoke a VX handle via IPC.  The IPC type must be VXIPC_CALL.
int vx_call(vxkey *key, int ctl, struct vxcall *args);


// Get and/or modify the trust flag on a key.
// If 'newtrust' >= 0, sets the trust flag to 'newtrust'.
// Returns the old flag value in any case.
int vx_trust(vxkey *key, int newtrust);


// Release an arbitrary VX object referenced by a key.
void vx_free(vxkey *key);


// Set errno appropriately after receiving an unexpected reply from an IPC.
// If the reply uses the special VXERRCODE return code, set errno from s1;
// otherwise use the default error code EPROTO (Protocol Error).
#define vx_seterrno(rc, s1) \
	if (rc == VXERRCODE | VXIPC_RETURN | VXIPC_SEND0) \
		errno = s1; \
	else \
		errno = EPROTO;


#endif	// VX_IPC_H
