//
// High-level VX process control interface.
// Usable both in a host OS to create and control VX environments,
// and within a VX environment to create and control sub-processes.
//
#ifndef VX_PROC_H
#define VX_PROC_H

#include <sys/types.h>
#include <sys/signal.h>	// XX FreeBSD-4.9 header bug?
#include <ucontext.h>

#include <vx/vx.h>


typedef struct vxproc vxproc;


// Request/reply message structure for handling parent-calls
typedef struct vxpmsg {
	size_t		code;		// request/reply code (rAX)
	size_t		val1;		// simple register value 1 (rSI)
	size_t		val2;		// simple register value 2 (rDI)
	union {
		void *	data;		// byte-string data buffer (rBX)
		size_t	valB;		// ...or simple register value B
	};
	union {
		size_t	datalen;	// byte-string data length (rDX)
		size_t	valL;		// ...or simple register value S
	};
} vxpmsg;


// Create a new VX process, initially empty.
vxproc *vxproc_new(void);

// Set permissions on a memory region within a VX process.
int vxproc_setperm(vxproc *proc, intptr_t addr, size_t size, int perm);

// Check for specified permissions on a memory region.
// Returns nonzero if entire region has the given permissions, zero if not.
// If not, and if out_faultva != NULL, sets *out_faultva to
// the lowest virtual address in the region on which permission check fails.
int vxproc_checkperm(vxproc *p, size_t va, size_t len, int perm,
		size_t *out_faultva);

// Map an address range of a VX process into host memory for direct access.
// 'perm' specifies the required access permissions for the region,
// and is checked against the permissions active within the VX process,
// returning NULL with errno=EACCES if the permission check fails.
// Client must call vxproc_unmaprange before mapping a different region
// or performing any other vxproc operations (such as running the environment).
void *vxproc_maprange(vxproc *p, size_t va, size_t len, int perm);
void vxproc_unmaprange(vxproc *p);

// Read or write an address region within a VX subprocess,
// disregarding the page permissions for that region.
int vxproc_read(vxproc *proc, intptr_t addr, void *buf, size_t size);
int vxproc_write(vxproc *proc, const void *data, intptr_t addr, size_t size);

// Load an executable image into a VX environment,
// initializing the environment's stack and register state as appropriate.
// Uses the supplied read callback to read portions of the ELF executable.
int vxproc_load(vxproc *proc,
		ssize_t (*readcb)(void *data, off_t ofs,
				void *buf, size_t size),
		void *cbdata,
		const char *const argv [], const char *const envp []);

// Load an executable image from host memory into a VX process.
int vxproc_loadmem(vxproc *proc, const void *exe, size_t exesize,
		const char *const argv [], const char *const envp []);

// Load a VX executable from a file on the host into a VX process.
int vxproc_loadfile(vxproc *proc, const char *filename,
		const char *const argv [], const char *const envp []);

// Set the parent-call handler to service this process's parent-calls.
typedef int vxpcallhandler(vxproc *proc, vxpmsg *msg, void *pcalldata);
void vxproc_setpcall(vxproc *proc, vxpcallhandler *handler, void *pcalldata,
			size_t maxmsglen);

// Set up UNIX signal handlers to catch signals caused by VX code.
// You must either do this before calling vxproc_run() in a given thread,
// or must install your own SIGSEGV, SIGBUS, and SIGFPE handlers
// that call vxproc_sighandler (below) to handle VX process signals.
int vxproc_siginit(void);

// Clients that do not use vxproc_siginit() above must call
// vxproc_sighandler() from their SIGSEGV, SIGBUS, and SIGFPE handlers
// These handlers must run on an alternate signal stack.
// This function returns nonzero if the signal was recognized as being
// generated by the VX environment and has been handled appropriately,
// or nonzero if the signal was unrecognized and has not been handled.
int vxproc_sighandler(int signo, mcontext_t *mcontext);

// Run the VX process until an exception or other terminating event occurs.
int vxproc_run(vxproc *proc);

vxproc *vxproc_dup(vxproc *proc);

void vxproc_free(vxproc *proc);


#endif	// VX_PROC_H
