//
// Basic definitions for VX environment
//
#ifndef VX_H
#define VX_H

#include <inttypes.h>


struct vxproc;


// VX32 processes have a 30-bit address space (1GB).
#define VXMEMSHIFT	30
#define VXMEMSIZE	(1 << VXMEMSHIFT)

// VX memory access permissions have a granularity of 4KB pages.
#define VXPAGESHIFT	12
#define VXPAGESIZE	(1 << VXPAGESHIFT)		// 4KB
#define VXPAGETRUNC(v)	((uint32_t)(v) & ~(VXPAGESIZE-1))
#define VXPAGEROUND(v)	VXPAGETRUNC((uint32_t)(v) + (VXPAGESIZE-1))

// Page permission bits
#define VXPERM_READ	0x01
#define VXPERM_WRITE	0x02
#define VXPERM_EXEC	0x04


// VX trap code categories
#define VXTRAP_CPU		0x100	// General CPU-detected exception
#define VXTRAP_SOFT		0x200	// Software interrupt (INT $n)

// Error codes indicating traps within the VX environment
#define VXTRAP_DIVIDE		0x100	// Integer divide by zero
#define VXTRAP_BREAKPOINT	0x103	// INT3 (breakpoint) instruction
#define VXTRAP_INVALID		0x106	// Invalid instruction opcode
#define VXTRAP_PAGEFAULT	0x10e	// Reference to inaccessible page
#define VXTRAP_ALIGN		0x111	// Misaligned memory reference
#define VXTRAP_FLOAT		0x113	// SIMD floating-point exception

// Traps relating to system calls
#define VXTRAP_SYS		0x180	// Unrecognized system call
#define VXTRAP_INVARG		0x181	// Invalid argument to system call


// General emulator error codes
#define VXERR_BADEXEC		1	// Bad executable chunk in vx_load()


// VX environment system calls (syscalls).
// These system calls represent architectural services
// that support VX code in controlling and organizing
// its own internal computation,
// such as manipulating its own address space
// and creating child environments for nested computations.
// Primitive system calls do not provide any external I/O capabilities.
// As such, the VX environment generally services all
// primitive system calls directly regardless of nesting level,
// except where part of the system call interface may be disabled
// for architecture reduction or standardization purposes.
//
// System calls use the x86/vx32 SYSCALL instruction,
// with the system call code in the low 8 bits of EAX,
// and the other registers used according to the system call.
//
#define VXSYS_PCALL		0x00	// Parent call - see below
#define VXSYS_SETPERM		0x01	// Set page permissions uniformly
#define VXSYS_PERM		0x02	// Set or get page permissions


// Parent-calls (pcalls) provide the standard way
// for code in a VX environment to communicate with its parent environment,
// including performing any external I/O that it may require.
// This facility effectively serves as a simple synchronous IPC mechanism
// in which the target process is implicitly always "the parent",
// and thus communication always conforms to environment nesting relationships.
//
// A VX environment invokes a parent-call via the VXSYS_PCALL syscall,
// placing VXSYS_PCALL in the low 8 bits of EAX,
// call flags in the next 4 bits of EAX,
// a parent-call code in the upper 20 bits of EAX,
// and other registers used for passing call and return arguments.
//
#define VXPC_SENDBUF		0x100	// Send data buffer in ebx, len=edx

// Pcall codes are specific to a particular use of the VX environment,
// but some standardized codes are defined below
// to provide a uniform interface to basic POSIX-ish services
// common to many different VX environment variations.
//
#define VXPC_EXIT		0x01000	// Exit with status code
#define VXPC_READ		0x02000	// Read from a file
#define VXPC_WRITE		0x03100	// Write to a file


unsigned vx_error();

const char *vx_errmsg();


#endif	// VX_H
