// Low-level control functions for 32-bit VX32 environments
#ifndef VX32_H
#define VX32_H

#include <inttypes.h>


struct vxproc;


// EFLAGS bits valid for VX32 code
#define VX32_OF		0x0800		// Overflow Flag
#define VX32_DF		0x0400		// Direction Flag
#define VX32_SF		0x0080		// Sign Flag
#define VX32_ZF		0x0040		// Zero Flag
#define VX32_CF		0x0001		// Carry Flag

// All EFLAGS bits valid for VX32 code
#define VX32_EFLAGS	(VX32_OF | VX32_DF | VX32_SF | VX32_ZF | VX32_CF)


// Union describing a 128-bit XMM vector register.
// Contents are always stored in little-endian byte order.
// Union members other than i8 should ONLY be accessed by the host
// if the host is also little-endian, or does manual byte swapping.
union vxvec {
	int8_t		i8[16];
	int16_t		i16[8];
	int32_t		i32[4];
	int32_t		i64[2];
	float		f32[4];
	float		f64[2];
} __attribute__ ((aligned (16)));;


// VX32 register state frame structure,
// used in vx_getreg() and vx_setreg() calls.
struct vx32regs {

	// General-purpose registers, stored in host byte order.
	uint32_t	eax;
	uint32_t	ecx;
	uint32_t	edx;
	uint32_t	ebx;
	uint32_t	esp;
	uint32_t	ebp;
	uint32_t	esi;
	uint32_t	edi;

	// Special registers, host byte order.
	uint32_t	eip;
	uint32_t	eflags;
	uint32_t	mxcsr;
	uint32_t	reserved;

	// XMM vector registers, little-endian byte order.
	union vxvec	xmm[8];
};


// VX32 primary code chunk header
struct vx32codehdr {
	uint32_t	signature;	// Magic value: VX32_CODESIG
	uint32_t	flags;		// Load flags
	uint32_t	loadlim;	// Load VA of end of primary code chunk
	uint32_t	entry;		// VA of entrypoint into code
};
#define VX32_CODESIG	0x32335856	// 'VX32' (little-endian)



// Size of the initial stack the vx32 loader sets up for a new process,
// which grows downward from the lowest virtual address used by the image.
// The process is free to grow or move its stack after startup, however.
#define VX32_STACK	(64*1024)


// Maximum size of a process's argument and environment arrays on the stack.
// Corresponds to _POSIX_ARG_MAX, the POSIX-specified lower bound.
#define VX32_ARG_MAX	(4*1024)



// Load a VX32 primary code chunk into a VX process,
// and initialize the vxproc's register state appropriately.
int vx32_load(struct vxproc *proc,
		ssize_t (*readcb)(void *data, off_t ofs,
				void *buf, size_t size),
		void *cbdata,
		const char *const argv [], const char *const envp []);

// Examine or modify the current non-SSE register state of a VX32 process
void vx32_getregs(struct vxproc *proc, struct vx32regs *regs);
void vx32_setregs(struct vxproc *proc, const struct vx32regs *regs);

// Examine or modify the current SSE register state of a VX32 process
void vx32_getxmm(struct vxproc *proc, struct vx32regs *regs);
void vx32_setxmm(struct vxproc *proc, const struct vx32regs *regs);

// Examine or modify the complete register state of a VX32 process
void vx32_getstate(struct vxproc *proc, struct vx32regs *regs);
void vx32_setstate(struct vxproc *proc, const struct vx32regs *regs);


#endif	// VX32_H
