/* Code specific to x86 hosts running FreeBSD */

#include <string.h>
#include <signal.h>
#include <assert.h>

#include <machine/segments.h>
#include <machine/sysarch.h>

#include "env/env.h"
#include "emu.h"
#include "os.h"


// First LDT selector number to use for our translator segments.
// Avoid stepping on the low NLDT selectors defined by the kernel.
// XXX I have the impression some versions of FreeBSD TLS
// may also use variable LDT segments; should avoid stepping on them...
#define SELOFS	((32*8) + 4 + 3)	// 4=LDT, 3=RPL


static void setbase(struct segment_descriptor *desc, unsigned long base)
{
	desc->sd_lobase = base & 0xffffff;
	desc->sd_hibase = base >> 24;
}

static void setlimit(struct segment_descriptor *desc, unsigned long limit)
{
	desc->sd_lolimit = limit & 0xffff;
	desc->sd_hilimit = limit >> 16;
}


// Set up LDT segments needed by the instruction translater
// whenever a vx32 process's memory is mapped.
int vxemu_map(struct vxemu *emu)
{
	struct vxproc *vxp = emu->proc;

	assert(SELOFS >= NLDT*8);
	emu->datasel = SELOFS + vxp->vxpno * 16;
	emu->emusel = emu->datasel + 8;

	// Set up the process's data segment selector (for DS,ES,SS).
	union descriptor desc;
	memset(&desc, 0, sizeof(desc));
	setbase(&desc.sd, (unsigned long)vxp->data);
	setlimit(&desc.sd, (vxp->size - 1) >> VXPAGESHIFT);
	desc.sd.sd_type = SDT_MEMRWA;
	desc.sd.sd_dpl = 3;
	desc.sd.sd_p = 1;
	desc.sd.sd_def32 = 1;
	desc.sd.sd_gran = 1;
	if (i386_set_ldt(emu->datasel >> 3, &desc, 1) < 0)
		return -1;

	// Set up the process's vxemu segment selector (for FS).
	setbase(&desc.sd, (unsigned long)emu);
	setlimit(&desc.sd, (VXCODEBUFSIZE - 1) >> VXPAGESHIFT);
	if (i386_set_ldt(emu->emusel >> 3, &desc, 1) < 0)
		return -1;

#ifdef __x86_64
	// Set up 32-bit mode code and data segments (not vxproc-specific),
	// giving access to the full low 32-bit of linear address space.
	// The code segment is necessary to get into 32-bit compatibility mode;
	// the data segment is needed because Linux for x86-64
	// doesn't give 64-bit processes a "real" data segment by default
	// but instead just loads zero into the data segment selectors!
	XXX adapt to FreeBSD
	emu->runptr.sel = FLATCODE;
	desc.entry_number = emu->runptr.sel / 8;
	desc.base_addr = 0;
	desc.limit = 0xfffff;
	desc.contents = MODIFY_LDT_CONTENTS_CODE;
	if (modify_ldt(1, &desc, sizeof(desc)) < 0)
		return -1;

	desc.entry_number = FLATDATA / 8;
	desc.contents = MODIFY_LDT_CONTENTS_DATA;
	if (modify_ldt(1, &desc, sizeof(desc)) < 0)
		return -1;

	// Set up a far return vector in emu->retptr
	// for getting back into 64-bit long mode.
	extern void vxrun_return();
	asm volatile("movw %%cs,%0" : "=r" (emu->retptr.sel));
	emu->retptr.ofs = (uint32_t)(intptr_t)vxrun_return;
#endif

	return 0;
}

// Note: it's not generally safe to call printf from here,
// even just for debugging, because we're running on a small signal stack.
int vxproc_sighandler(int signo, mcontext_t *ctx)
{
	// See if we're in the context of the translator,
	// by testing whether DS is one of our LDT selectors.
	// XXX the process could use other LDT selectors.
	if (ctx->mc_ds < SELOFS) {
		return 0;
	}

#ifdef i386
	size_t trapeip = ctx->mc_eip;
#else
	size_t trapeip = ctx->mc_rip;
#endif

	// Use our vxemu segment to find our current vxproc and vxemu struct.
	vxproc *vxp;
	asm("mov %"VSEGSTR":%1,%0"
		: "=r" (vxp)
		: "m" (((vxemu*)0)->proc));
	vxemu *emu = vxp->emu;

	// Get back our regular host segment register state,
	// so that thread-local storage and such works.
	vxrun_cleanup(emu);

	vxemu_generic_sighandler(emu, trapeip, ctx->mc_trapno, ctx->mc_err);
}

