/* Code specific to x86 hosts running Linux */

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <assert.h>
#include <ucontext.h>
#include <asm/ldt.h>

#include "env/env.h"
#include "emu.h"
#include "os.h"


int modify_ldt(int func, void *ptr, unsigned long bytecount);


// Set up LDT segments needed by the instruction translater
// whenever a vx32 process's memory is mapped.
int vxemu_map(struct vxemu *emu)
{
	struct vxproc *vxp = emu->proc;
	emu->datasel = vxp->vxpno * 16 + 16 + 4 + 3;	// 4=LDT, 3=RPL
	emu->emusel = emu->datasel + 8;

	// Set up the process's data segment selector (for DS,ES,SS).
	struct user_desc desc;
	memset(&desc, 0, sizeof(desc));
	desc.entry_number = emu->datasel / 8;
	desc.base_addr = (unsigned long)vxp->data;
	desc.limit = (vxp->size - 1) >> VXPAGESHIFT;
	desc.seg_32bit = 1;
	desc.contents = MODIFY_LDT_CONTENTS_DATA;
	desc.read_exec_only = 0;
	desc.limit_in_pages = 1;
	desc.seg_not_present = 0;
	desc.useable = 1;
	if (modify_ldt(1, &desc, sizeof(desc)) < 0)
		return -1;

	// Set up the process's vxemu segment selector (for FS).
	desc.entry_number = emu->emusel / 8;
	desc.base_addr = (unsigned long)emu;
	desc.limit = (VXCODEBUFSIZE - 1) >> VXPAGESHIFT;
	if (modify_ldt(1, &desc, sizeof(desc)) < 0)
		return -1;

#ifdef __x86_64
	// Set up 32-bit mode code and data segments (not vxproc-specific),
	// giving access to the full low 32-bit of linear address space.
	// The code segment is necessary to get into 32-bit compatibility mode;
	// the data segment is needed because Linux for x86-64
	// doesn't give 64-bit processes a "real" data segment by default
	// but instead just loads zero into the data segment selectors!
	emu->runptr.sel = FLATCODE;
	desc.entry_number = emu->runptr.sel / 8;
	desc.base_addr = 0;
	desc.limit = 0xfffff;
	desc.contents = MODIFY_LDT_CONTENTS_CODE;
	if (modify_ldt(1, &desc, sizeof(desc)) < 0)
		return -1;

	desc.entry_number = FLATDATA / 8;
	desc.contents = MODIFY_LDT_CONTENTS_DATA;
	if (modify_ldt(1, &desc, sizeof(desc)) < 0)
		return -1;

	// Set up a far return vector in emu->retptr
	// for getting back into 64-bit long mode.
	extern void vxrun_return();
	asm volatile("movw %%cs,%0" : "=r" (emu->retptr.sel));
	emu->retptr.ofs = (uint32_t)(intptr_t)vxrun_return;
#endif

	return 0;
}

static void dumpsigcontext(struct sigcontext *ctx)
{
#ifdef i386
	printf(
		"eax %08lx  ebx %08lx\necx %08lx  edx %08lx  "
		"rsi %08lx  rdi %08lx\nrbp %08lx  rsp %08lx\n"
		"eip %08lx  efl %08lx  cs %04x\n"
		"err %08lx  trapno %08lx  cr2 %08lx\n",
		ctx->eax, ctx->ebx, ctx->ecx, ctx->edx,
		ctx->esi, ctx->edi, ctx->ebp, ctx->esp,
		ctx->eip, ctx->eflags, ctx->cs,
		ctx->err, ctx->trapno, ctx->cr2);
#else
	printf(
		"rax %016lx  rbx %016lx\nrcx %016lx  rdx %016lx\n"
		"rsi %016lx  rdi %016lx\nrbp %016lx  rsp %016lx\n"
		"r8  %016lx  r9  %016lx\nr10 %016lx  r11 %016lx\n"
		"r12 %016lx  r13 %016lx\nr14 %016lx  r15 %016lx\n"
		"rip %016lx  efl %016lx  cs %04x  ss %04x\n"
		"err %016lx  trapno %016lx  cr2 %016lx\n",
		ctx->rax, ctx->rbx, ctx->rcx, ctx->rdx,
		ctx->rsi, ctx->rdi, ctx->rbp, ctx->rsp,
		ctx->r8, ctx->r9, ctx->r10, ctx->r11,
		ctx->r12, ctx->r13, ctx->r14, ctx->r15,
		ctx->rip, ctx->eflags, ctx->cs, ctx->__pad0,
		ctx->err, ctx->trapno, ctx->cr2);
#endif
}

// Note: it's not generally safe to call printf from here,
// even just for debugging, because we're running on a small signal stack.
int vxproc_sighandler(int signo, mcontext_t *mctx)
{
	// same layout, and sigcontext is more convenient...
	struct sigcontext *ctx = (struct sigcontext*)ctx;

	// See if we're in the context of the translator,
	// by testing whether CS is our 32-bit code segment.
	// XXX this only works for x86-64!!! FIX!!!
	if (ctx->cs != FLATCODE) {
		return 0;
	}

#ifdef i386
	size_t trapeip = ctx->eip;
#else
	size_t trapeip = ctx->rip;
#endif

	// Use FS to find our current vxproc and vxemu.
	vxproc *vxp;
	asm("mov %"VSEGSTR":%1,%0"
		: "=r" (vxp)
		: "m" (((vxemu*)0)->proc));
	vxemu *emu = vxp->emu;

	// Get back our regular host segment register state,
	// so that thread-local storage and such works.
	vxrun_cleanup(emu);

#if 0
fprintf(stderr, "FOO\n");
dumpsegs("sighandler");
dumpsigcontext(ctx);
dumpsegs("sighandler");
fflush(stdout);
fflush(stderr);
#endif

	vxemu_generic_sighandler(emu, trapeip, ctx->trapno, ctx->err);
}

