
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <signal.h>

#include "env/env.h"
#include "emu.h"
#include "os.h"

// After a trap occurs in the translated code,
// find the corresponding EIP in the original vx32 code,
// and fix up the condition codes if necessary.
// Assumes the trap-time registers
// have already been saved back into the vxemu state.
// codeofs = offset of trap-time EIP in codebuf.
static void trapfix(struct vxemu *emu, uint32_t codeofs)
{
	// Binary search for the translated chunk in which the trap occurred.
	// NB: frags appear in the opposite order as the code itself.
	uint32_t *codetab = emu->codetab;
	unsigned lofrag = 0;
	unsigned hifrag = (uint32_t*)emu->codetop - codetab - 1;
	while (hifrag > lofrag) {
		unsigned midfrag = (lofrag + hifrag) / 2;
		uint32_t midofs = codetab[midfrag];
		if (codeofs >= midofs)
			hifrag = midfrag;
		else
			lofrag = midfrag + 1;
	}
	struct vxfrag *frag = emu->codebuf + codetab[lofrag];

	// Now binary search through this fragment's instruction table
	// for the actual faulting translated instruction,
	// and compute the corresponding EIP in the original vx32 code.
	unsigned ninsn = frag->ninsn;
	unsigned insnofs = codeofs -
			((void*)FRAGCODE(frag) - emu->codebuf);
	unsigned loinsn = 0;
	unsigned hiinsn = ninsn - 1;
	while (hiinsn > loinsn) {
		unsigned midinsn = (loinsn + hiinsn + 1) / 2;
		unsigned midofs = frag->insn[midinsn].dstofs;
		if (insnofs >= midofs)
			loinsn = midinsn;
		else
			hiinsn = midinsn - 1;
	}
	struct vxinsn *insn = &frag->insn[loinsn];

	// Find the corresponding EIP in the original code
	emu->eip = frag->eip + insn->srcofs;

#if 0
	// Fix up the condition codes if necessary
	if (insn->flags & INSN_FIXCC) {
		// XXX
		assert(0);
	}
#endif
}

// OS-independent signal recovery code.
// Assumes OS-specific code has already verified
// that the signal was caused by translated vx32 code.
// Doesn't return; always longjmps back to the translater.
//
// Note: it's not generally safe to call printf from here,
// even just for debugging, because we're running on a small signal stack.
//
void vxemu_generic_sighandler(vxemu *emu, unsigned trapeip,
				unsigned trapno, unsigned err)
{
	// Binary search for the translated chunk in which the trap occurred.
	// NB: frags appear in the opposite order as the code itself.
	uint32_t *codetab = emu->codetab;
	unsigned lofrag = 0;
	unsigned hifrag = (uint32_t*)emu->codetop - codetab - 1;
	while (hifrag > lofrag) {
		unsigned midfrag = (lofrag + hifrag) / 2;
		uint32_t mideip = codetab[midfrag];
		if (trapeip >= mideip)
			hifrag = midfrag;
		else
			lofrag = midfrag + 1;
	}
	struct vxfrag *frag = (vxfrag*)(intptr_t)codetab[lofrag];

	// Now binary search through this fragment's instruction table
	// for the actual faulting translated instruction,
	// and compute the corresponding EIP in the original vx32 code.
	unsigned ninsn = frag->ninsn;
	unsigned insnofs = trapeip - (uint32_t)(intptr_t)FRAGCODE(frag);
	unsigned loinsn = 0;
	unsigned hiinsn = ninsn - 1;
	while (hiinsn > loinsn) {
		unsigned midinsn = (loinsn + hiinsn + 1) / 2;
		unsigned midofs = frag->insn[midinsn].dstofs;
		if (insnofs >= midofs)
			loinsn = midinsn;
		else
			hiinsn = midinsn - 1;
	}
	struct vxinsn *insn = &frag->insn[loinsn];

	// Find the corresponding EIP in the original code
	emu->eip = frag->eip + insn->srcofs;

	// XXX fixup registers as necessary

	emu->trapno = trapno;
	emu->traperr = err;
	emu->trapva = 0;	// XXX

	longjmp(*emu->trapenv, 1);
}

